# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import sys
import unittest

import libcloud.pricing

PRICING_FILE_PATH = os.path.join(os.path.dirname(__file__), 'pricing_test.json')


class PricingTestCase(unittest.TestCase):
    def setUp(self):
        super(PricingTestCase, self).setUp()

        libcloud.pricing.PRICING_DATA = {"compute": {}, "storage": {}}

    def test_get_pricing_success(self):
        self.assertFalse('foo' in libcloud.pricing.PRICING_DATA['compute'])

        pricing = libcloud.pricing.get_pricing(driver_type='compute',
                                               driver_name='foo',
                                               pricing_file_path=PRICING_FILE_PATH)
        self.assertEqual(pricing['1'], 1.0)
        self.assertEqual(pricing['2'], 2.0)

        self.assertEqual(libcloud.pricing.PRICING_DATA['compute']['foo']['1'], 1.0)
        self.assertEqual(libcloud.pricing.PRICING_DATA['compute']['foo']['2'], 2.0)

    def test_get_pricing_invalid_file_path(self):
        try:
            libcloud.pricing.get_pricing(driver_type='compute', driver_name='bar',
                                         pricing_file_path='inexistent.json')
        except IOError:
            pass
        else:
            self.fail('Invalid pricing file path provided, but an exception was not'
                      ' thrown')

    def test_get_pricing_invalid_driver_type(self):
        try:
            libcloud.pricing.get_pricing(driver_type='invalid_type', driver_name='bar',
                                         pricing_file_path='inexistent.json')
        except AttributeError:
            pass
        else:
            self.fail('Invalid driver_type provided, but an exception was not'
                      ' thrown')

    def test_get_pricing_not_in_cache(self):
        try:
            libcloud.pricing.get_pricing(driver_type='compute', driver_name='inexistent',
                                         pricing_file_path=PRICING_FILE_PATH)
        except KeyError:
            pass
        else:
            self.fail('Invalid driver provided, but an exception was not'
                      ' thrown')

    def test_get_size_price(self):
        libcloud.pricing.PRICING_DATA['compute']['foo'] = {2: 2, '3': 3}
        price1 = libcloud.pricing.get_size_price(driver_type='compute',
                                                 driver_name='foo',
                                                 size_id=2)
        price2 = libcloud.pricing.get_size_price(driver_type='compute',
                                                 driver_name='foo',
                                                 size_id='3')
        self.assertEqual(price1, 2)
        self.assertEqual(price2, 3)

    def test_invalid_pricing_cache(self):
        libcloud.pricing.PRICING_DATA['compute']['foo'] = {2: 2}
        self.assertTrue('foo' in libcloud.pricing.PRICING_DATA['compute'])

        libcloud.pricing.invalidate_pricing_cache()
        self.assertFalse('foo' in libcloud.pricing.PRICING_DATA['compute'])

    def test_invalid_module_pricing_cache(self):
        libcloud.pricing.PRICING_DATA['compute']['foo'] = {1: 1}

        self.assertTrue('foo' in libcloud.pricing.PRICING_DATA['compute'])

        libcloud.pricing.invalidate_module_pricing_cache(driver_type='compute',
                                                         driver_name='foo')
        self.assertFalse('foo' in libcloud.pricing.PRICING_DATA['compute'])
        libcloud.pricing.invalidate_module_pricing_cache(driver_type='compute',
                                                         driver_name='foo1')

    def test_set_pricing(self):
        self.assertFalse('foo' in libcloud.pricing.PRICING_DATA['compute'])

        libcloud.pricing.set_pricing(driver_type='compute', driver_name='foo',
                                     pricing={'foo': 1})
        self.assertTrue('foo' in libcloud.pricing.PRICING_DATA['compute'])

    def test_get_pricing_data_caching(self):
        # Ensure we only cache pricing data in memory for requested drivers
        self.assertEqual(libcloud.pricing.PRICING_DATA["compute"], {})
        self.assertEqual(libcloud.pricing.PRICING_DATA["storage"], {})

        pricing = libcloud.pricing.get_pricing(driver_type='compute',
                                               driver_name='foo',
                                               pricing_file_path=PRICING_FILE_PATH)
        self.assertEqual(pricing['1'], 1.0)
        self.assertEqual(pricing['2'], 2.0)

        self.assertEqual(len(libcloud.pricing.PRICING_DATA["compute"]), 1)
        self.assertTrue("foo" in libcloud.pricing.PRICING_DATA["compute"])

        pricing = libcloud.pricing.get_pricing(driver_type='compute',
                                               driver_name='baz',
                                               pricing_file_path=PRICING_FILE_PATH)
        self.assertEqual(pricing['1'], 5.0)
        self.assertEqual(pricing['2'], 6.0)

        self.assertEqual(len(libcloud.pricing.PRICING_DATA["compute"]), 2)
        self.assertTrue("foo" in libcloud.pricing.PRICING_DATA["compute"])
        self.assertTrue("baz" in libcloud.pricing.PRICING_DATA["compute"])

    def test_get_pricing_data_module_level_variable_is_true(self):
        # Ensure we only cache pricing data in memory for requested drivers
        self.assertEqual(libcloud.pricing.PRICING_DATA["compute"], {})
        self.assertEqual(libcloud.pricing.PRICING_DATA["storage"], {})

        libcloud.pricing.CACHE_ALL_PRICING_DATA = True

        pricing = libcloud.pricing.get_pricing(driver_type='compute',
                                               driver_name='foo',
                                               pricing_file_path=PRICING_FILE_PATH,
                                               cache_all=False)
        self.assertEqual(pricing['1'], 1.0)
        self.assertEqual(pricing['2'], 2.0)

        self.assertEqual(len(libcloud.pricing.PRICING_DATA["compute"]), 3)
        self.assertTrue("foo" in libcloud.pricing.PRICING_DATA["compute"])
        self.assertTrue("bar" in libcloud.pricing.PRICING_DATA["compute"])
        self.assertTrue("baz" in libcloud.pricing.PRICING_DATA["compute"])

    def test_get_pricing_data_caching_cache_all(self):
        # Ensure we only cache pricing data in memory for requested drivers
        self.assertEqual(libcloud.pricing.PRICING_DATA["compute"], {})
        self.assertEqual(libcloud.pricing.PRICING_DATA["storage"], {})

        pricing = libcloud.pricing.get_pricing(driver_type='compute',
                                               driver_name='foo',
                                               pricing_file_path=PRICING_FILE_PATH,
                                               cache_all=True)
        self.assertEqual(pricing['1'], 1.0)
        self.assertEqual(pricing['2'], 2.0)

        self.assertEqual(len(libcloud.pricing.PRICING_DATA["compute"]), 3)
        self.assertTrue("foo" in libcloud.pricing.PRICING_DATA["compute"])
        self.assertTrue("bar" in libcloud.pricing.PRICING_DATA["compute"])
        self.assertTrue("baz" in libcloud.pricing.PRICING_DATA["compute"])


if __name__ == '__main__':
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Scaleway Driver
"""

import copy
try:
    import simplejson as json
except ImportError:
    import json

from libcloud.common.base import ConnectionUserAndKey, JsonResponse
from libcloud.common.types import ProviderError
from libcloud.compute.base import NodeDriver, NodeImage, Node, NodeSize
from libcloud.compute.base import NodeLocation
from libcloud.compute.base import StorageVolume, VolumeSnapshot, KeyPair
from libcloud.compute.providers import Provider
from libcloud.compute.types import NodeState, VolumeSnapshotState
from libcloud.utils.iso8601 import parse_date
from libcloud.utils.py3 import httplib

__all__ = [
    'ScalewayResponse',
    'ScalewayConnection',
    'ScalewayNodeDriver'
]

SCALEWAY_API_HOSTS = {
    'default': 'cp-par1.scaleway.com',
    'account': 'account.scaleway.com',
    'par1': 'cp-par1.scaleway.com',
    'ams1': 'cp-ams1.scaleway.com',
}

# The API doesn't give location info, so we provide it ourselves, instead.
SCALEWAY_LOCATION_DATA = [
    {'id': 'par1', 'name': 'Paris 1', 'country': 'FR'},
    {'id': 'ams1', 'name': 'Amsterdam 1', 'country': 'NL'},
]


class ScalewayResponse(JsonResponse):
    valid_response_codes = [httplib.OK, httplib.ACCEPTED,
                            httplib.CREATED, httplib.NO_CONTENT]

    def parse_error(self):
        return super(ScalewayResponse, self).parse_error()['message']

    def success(self):
        return self.status in self.valid_response_codes


class ScalewayConnection(ConnectionUserAndKey):
    """
    Connection class for the Scaleway driver.
    """

    host = SCALEWAY_API_HOSTS['default']
    allow_insecure = False
    responseCls = ScalewayResponse

    def request(self, action, params=None, data=None, headers=None,
                method='GET', raw=False, stream=False, region=None):
        if region:
            old_host = self.host
            self.host = SCALEWAY_API_HOSTS[region.id
                                           if isinstance(region, NodeLocation)
                                           else region]
            if not self.host == old_host:
                self.connect()

        return super(ScalewayConnection, self).request(action, params, data,
                                                       headers, method, raw,
                                                       stream)

    def _request_paged(self, action, params=None, data=None, headers=None,
                       method='GET', raw=False, stream=False, region=None):
        if params is None:
            params = {}

        if isinstance(params, dict):
            params['per_page'] = 100
        elif isinstance(params, list):
            params.append(('per_page', 100))  # pylint: disable=no-member

        results = self.request(action, params, data, headers,
                               method, raw, stream, region).object

        links = self.connection.getresponse().links
        while links and 'next' in links:
            next = self.request(links['next']['url'], data=data,
                                headers=headers, method=method,
                                raw=raw, stream=stream).object
            links = self.connection.getresponse().links
            merged = {root: child + next[root]
                      for root, child in list(results.items())}
            results = merged

        return results

    def add_default_headers(self, headers):
        """
        Add headers that are necessary for every request
        """
        headers['X-Auth-Token'] = self.key
        headers['Content-Type'] = 'application/json'
        return headers


def _to_lib_size(size):
    return int(size / 1000 / 1000 / 1000)


def _to_api_size(size):
    return int(size * 1000 * 1000 * 1000)


class ScalewayNodeDriver(NodeDriver):
    """
    Scaleway Node Driver Class

    This is the primary driver for interacting with Scaleway.  It contains all
    of the standard libcloud methods that Scaleway's API supports.
    """

    type = Provider.SCALEWAY
    connectionCls = ScalewayConnection
    name = 'Scaleway'
    website = 'https://www.scaleway.com/'

    SNAPSHOT_STATE_MAP = {
        'snapshotting': VolumeSnapshotState.CREATING,
        'available': VolumeSnapshotState.AVAILABLE,
        'error': VolumeSnapshotState.ERROR
    }

    def list_locations(self):
        """
        List data centers available.

        :return: list of node location objects
        :rtype: ``list`` of :class:`.NodeLocation`
        """
        return [NodeLocation(driver=self, **copy.deepcopy(location))
                for location in SCALEWAY_LOCATION_DATA]

    def list_sizes(self, region=None):
        """
        List available VM sizes.

        :param region: The region in which to list sizes
        (if None, use default region specified in __init__)
        :type region: :class:`.NodeLocation`

        :return: list of node size objects
        :rtype: ``list`` of :class:`.NodeSize`
        """
        response = self.connection._request_paged('/products/servers',
                                                  region=region)
        sizes = response['servers']

        response = self.connection._request_paged(
            '/products/servers/availability', region=region)
        availability = response['servers']

        return sorted([self._to_size(name, sizes[name], availability[name])
                       for name in sizes], key=lambda x: x.name)

    def _to_size(self, name, size, availability):
        min_disk = (_to_lib_size(size['volumes_constraint']['min_size'] or 0)
                    if size['volumes_constraint'] else 25)
        max_disk = (_to_lib_size(size['volumes_constraint']['max_size'] or 0)
                    if size['volumes_constraint'] else min_disk)

        extra = {
            'cores': size['ncpus'],
            'monthly': size['monthly_price'],
            'arch': size['arch'],
            'baremetal': size['baremetal'],
            'availability': availability['availability'],
            'max_disk': max_disk,
            'internal_bandwidth': int(
                (size['network']['sum_internal_bandwidth'] or 0) /
                (1024 * 1024)),
            'ipv6': size['network']['ipv6_support'],
            'alt_names': size['alt_names'],
        }

        return NodeSize(id=name,
                        name=name,
                        ram=int((size['ram'] or 0) / (1024 * 1024)),
                        disk=min_disk,
                        bandwidth=int(
                            (size['network']['sum_internet_bandwidth'] or 0) /
                            (1024 * 1024)),
                        price=size['hourly_price'],
                        driver=self,
                        extra=extra)

    def list_images(self, region=None):
        """
        List available VM images.

        :param region: The region in which to list images
        (if None, use default region specified in __init__)
        :type region: :class:`.NodeLocation`

        :return: list of image objects
        :rtype: ``list`` of :class:`.NodeImage`
        """
        response = self.connection._request_paged('/images', region=region)
        images = response['images']
        return [self._to_image(image) for image in images]

    def create_image(self, node, name, region=None):
        """
        Create a VM image from an existing node's root volume.

        :param node: The node from which to create the image
        :type node: :class:`.Node`

        :param name: The name to give the image
        :type name: ``str``

        :param region: The region in which to create the image
        (if None, use default region specified in __init__)
        :type region: :class:`.NodeLocation`

        :return: the newly created image object
        :rtype: :class:`.NodeImage`
        """
        data = {
            'organization': self.key,
            'name': name,
            'arch': node.extra['arch'],
            'root_volume': node.extra['volumes']['0']['id']
        }
        response = self.connection.request('/images', data=json.dumps(data),
                                           region=region,
                                           method='POST')
        image = response.object['image']
        return self._to_image(image)

    def delete_image(self, node_image, region=None):
        """
        Delete a VM image.

        :param node_image: The image to delete
        :type node_image: :class:`.NodeImage`

        :param region: The region in which to find/delete the image
        (if None, use default region specified in __init__)
        :type region: :class:`.NodeLocation`

        :return: True if the image was deleted, otherwise False
        :rtype: ``bool``
        """
        return self.connection.request('/images/%s' % node_image.id,
                                       region=region,
                                       method='DELETE').success()

    def get_image(self, image_id, region=None):
        """
        Retrieve a specific VM image.

        :param image_id: The id of the image to retrieve
        :type image_id: ``int``

        :param region: The region in which to create the image
        (if None, use default region specified in __init__)
        :type region: :class:`.NodeLocation`

        :return: the requested image object
        :rtype: :class:`.NodeImage`
        """
        response = self.connection.request('/images/%s' % image_id,
                                           region=region)
        image = response.object['image']
        return self._to_image(image)

    def _to_image(self, image):
        extra = {
            'arch': image['arch'],
            'size': _to_lib_size(image.get('root_volume', {})
                                      .get('size', 0)) or 50,
            'creation_date': parse_date(image['creation_date']),
            'modification_date': parse_date(image['modification_date']),
            'organization': image['organization'],
        }
        return NodeImage(id=image['id'],
                         name=image['name'],
                         driver=self,
                         extra=extra)

    def list_nodes(self, region=None):
        """
        List all nodes.

        :param region: The region in which to look for nodes
        (if None, use default region specified in __init__)
        :type region: :class:`.NodeLocation`

        :return: list of node objects
        :rtype: ``list`` of :class:`.Node`
        """
        response = self.connection._request_paged('/servers', region=region)
        servers = response['servers']
        return [self._to_node(server) for server in servers]

    def _to_node(self, server):
        public_ip = server['public_ip']
        private_ip = server['private_ip']
        location = server['location'] or {}
        return Node(id=server['id'],
                    name=server['name'],
                    state=NodeState.fromstring(server['state']),
                    public_ips=[public_ip['address']] if public_ip else [],
                    private_ips=[private_ip] if private_ip else [],
                    driver=self,
                    extra={'volumes': server['volumes'],
                           'tags': server['tags'],
                           'arch': server['arch'],
                           'organization': server['organization'],
                           'region': location.get('zone_id', 'par1')},
                    created_at=parse_date(server['creation_date']))

    def create_node(self, name, size, image, ex_volumes=None, ex_tags=None,
                    region=None):
        """
        Create a new node.

        :param name: The name to give the node
        :type name: ``str``

        :param size: The size of node to create
        :type size: :class:`.NodeSize`

        :param image: The image to create the node with
        :type image: :class:`.NodeImage`

        :param ex_volumes: Additional volumes to create the node with
        :type ex_volumes: ``dict`` of :class:`.StorageVolume`s

        :param ex_tags: Tags to assign to the node
        :type ex_tags: ``list`` of ``str``

        :param region: The region in which to create the node
        (if None, use default region specified in __init__)
        :type region: :class:`.NodeLocation`

        :return: the newly created node object
        :rtype: :class:`.Node`
        """
        data = {
            'name': name,
            'organization': self.key,
            'image': image.id,
            'volumes': ex_volumes or {},
            'commercial_type': size.id,
            'tags': ex_tags or []
        }

        allocate_space = image.extra.get('size', 50)
        for volume in data['volumes']:
            allocate_space += _to_lib_size(volume['size'])

        while allocate_space < size.disk:
            if size.disk - allocate_space > 150:
                bump = 150
            else:
                bump = size.disk - allocate_space

            vol_num = len(data['volumes']) + 1
            data['volumes'][str(vol_num)] = {
                "name": "%s-%d" % (name, vol_num),
                "organization": self.key,
                "size": _to_api_size(bump),
                "volume_type": "l_ssd"
            }
            allocate_space += bump

        if allocate_space > size.extra.get('max_disk', size.disk):
            range = ("of %dGB" % size.disk
                     if size.extra.get('max_disk', size.disk) == size.disk else
                     "between %dGB and %dGB" %
                     (size.extra.get('max_disk', size.disk), size.disk))
            raise ProviderError(
                value=("%s only supports a total volume size %s; tried %dGB" %
                       (size.id, range, allocate_space)),
                http_code=400, driver=self)

        response = self.connection.request('/servers', data=json.dumps(data),
                                           region=region, method='POST')
        server = response.object['server']
        node = self._to_node(server)
        node.extra['region'] = (region.id if isinstance(region, NodeLocation)
                                else region) or 'par1'

        # Scaleway doesn't start servers by default, let's do it
        self._action(node.id, 'poweron')

        return node

    def _action(self, server_id, action, region=None):
        return self.connection.request('/servers/%s/action' % server_id,
                                       region=region,
                                       data=json.dumps({'action': action}),
                                       method='POST').success()

    def reboot_node(self, node):
        """
        Reboot a node.

        :param node: The node to be rebooted
        :type node: :class:`Node`

        :return: True if the reboot was successful, otherwise False
        :rtype: ``bool``
        """
        return self._action(node.id, 'reboot')

    def destroy_node(self, node):
        """
        Destroy a node.

        :param node: The node to be destroyed
        :type node: :class:`Node`

        :return: True if the destroy was successful, otherwise False
        :rtype: ``bool``
        """
        return self._action(node.id, 'terminate')

    def list_volumes(self, region=None):
        """
        Return a list of volumes.

        :param region: The region in which to look for volumes
        (if None, use default region specified in __init__)
        :type region: :class:`.NodeLocation`

        :return: A list of volume objects.
        :rtype: ``list`` of :class:`StorageVolume`
        """
        response = self.connection._request_paged('/volumes', region=region)
        volumes = response['volumes']
        return [self._to_volume(volume) for volume in volumes]

    def _to_volume(self, volume):
        extra = {
            'organization': volume['organization'],
            'volume_type': volume['volume_type'],
            'creation_date': parse_date(volume['creation_date']),
            'modification_date': parse_date(volume['modification_date']),
        }
        return StorageVolume(id=volume['id'],
                             name=volume['name'],
                             size=_to_lib_size(volume['size']),
                             driver=self,
                             extra=extra)

    def list_volume_snapshots(self, volume, region=None):
        """
        List snapshots for a storage volume.

        @inherits :class:`NodeDriver.list_volume_snapshots`

        :param region: The region in which to look for snapshots
        (if None, use default region specified in __init__)
        :type region: :class:`.NodeLocation`
        """
        response = self.connection._request_paged('/snapshots', region=region)
        snapshots = filter(lambda s: s['base_volume']['id'] == volume.id,
                           response['snapshots'])
        return [self._to_snapshot(snapshot) for snapshot in snapshots]

    def _to_snapshot(self, snapshot):
        state = self.SNAPSHOT_STATE_MAP.get(snapshot['state'],
                                            VolumeSnapshotState.UNKNOWN)
        extra = {
            'organization': snapshot['organization'],
            'volume_type': snapshot['volume_type'],
        }
        return VolumeSnapshot(id=snapshot['id'],
                              driver=self,
                              size=_to_lib_size(snapshot['size']),
                              created=parse_date(snapshot['creation_date']),
                              state=state,
                              extra=extra)

    def create_volume(self, size, name, region=None):
        """
        Create a new volume.

        :param size: Size of volume in gigabytes.
        :type size: ``int``

        :param name: Name of the volume to be created.
        :type name: ``str``

        :param region: The region in which to create the volume
        (if None, use default region specified in __init__)
        :type region: :class:`.NodeLocation`

        :return: The newly created volume.
        :rtype: :class:`StorageVolume`
        """
        data = {
            'name': name,
            'organization': self.key,
            'volume_type': 'l_ssd',
            'size': _to_api_size(size)
        }
        response = self.connection.request('/volumes',
                                           region=region,
                                           data=json.dumps(data),
                                           method='POST')
        volume = response.object['volume']
        return self._to_volume(volume)

    def create_volume_snapshot(self, volume, name, region=None):
        """
        Create snapshot from volume.

        :param volume: The volume to create a snapshot from
        :type volume: :class`StorageVolume`

        :param name: The name to give the snapshot
        :type name: ``str``

        :param region: The region in which to create the snapshot
        (if None, use default region specified in __init__)
        :type region: :class:`.NodeLocation`

        :return: The newly created snapshot.
        :rtype: :class:`VolumeSnapshot`
        """
        data = {
            'name': name,
            'organization': self.key,
            'volume_id': volume.id
        }
        response = self.connection.request('/snapshots',
                                           region=region,
                                           data=json.dumps(data),
                                           method='POST')
        snapshot = response.object['snapshot']
        return self._to_snapshot(snapshot)

    def destroy_volume(self, volume, region=None):
        """
        Destroys a storage volume.

        :param volume: Volume to be destroyed
        :type volume: :class:`StorageVolume`

        :param region: The region in which to look for the volume
        (if None, use default region specified in __init__)
        :type region: :class:`.NodeLocation`

        :return: True if the destroy was successful, otherwise False
        :rtype: ``bool``
        """
        return self.connection.request('/volumes/%s' % volume.id,
                                       region=region,
                                       method='DELETE').success()

    def destroy_volume_snapshot(self, snapshot, region=None):
        """
        Dostroy a volume snapshot

        :param snapshot: volume snapshot to destroy
        :type snapshot: class:`VolumeSnapshot`

        :param region: The region in which to look for the snapshot
        (if None, use default region specified in __init__)
        :type region: :class:`.NodeLocation`

        :return: True if the destroy was successful, otherwise False
        :rtype: ``bool``
        """
        return self.connection.request('/snapshots/%s' % snapshot.id,
                                       region=region,
                                       method='DELETE').success()

    def list_key_pairs(self):
        """
        List all the available SSH keys.

        :return: Available SSH keys.
        :rtype: ``list`` of :class:`KeyPair`
        """
        response = self.connection.request('/users/%s' % (self._get_user_id()),
                                           region='account')
        keys = response.object['user']['ssh_public_keys']
        return [KeyPair(name=' '.join(key['key'].split(' ')[2:]),
                        public_key=' '.join(key['key'].split(' ')[:2]),
                        fingerprint=key['fingerprint'],
                        driver=self) for key in keys]

    def import_key_pair_from_string(self, name, key_material):
        """
        Import a new public key from string.

        :param name: Key pair name.
        :type name: ``str``

        :param key_material: Public key material.
        :type key_material: ``str``

        :return: Imported key pair object.
        :rtype: :class:`KeyPair`
        """
        new_key = KeyPair(name=name,
                          public_key=' '.join(key_material.split(' ')[:2]),
                          fingerprint=None,
                          driver=self)
        keys = [key for key in self.list_key_pairs() if not key.name == name]
        keys.append(new_key)
        return self._save_keys(keys)

    def delete_key_pair(self, key_pair):
        """
        Delete an existing key pair.

        :param key_pair: Key pair object.
        :type key_pair: :class:`KeyPair`

        :return:   True of False based on success of Keypair deletion
        :rtype:    ``bool``
        """
        keys = [key for key in self.list_key_pairs()
                if not key.name == key_pair.name]
        return self._save_keys(keys)

    def _get_user_id(self):
        response = self.connection.request('/tokens/%s' % self.secret,
                                           region='account')
        return response.object['token']['user_id']

    def _save_keys(self, keys):
        data = {
            'ssh_public_keys': [{'key': '%s %s' % (key.public_key, key.name)}
                                for key in keys]
        }
        response = self.connection.request('/users/%s' % (self._get_user_id()),
                                           region='account',
                                           method='PATCH',
                                           data=json.dumps(data))
        return response.success()



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from base64 import b64encode

from libcloud.utils.py3 import b
from libcloud.utils.py3 import httplib
from libcloud.common.base import ConnectionUserAndKey, JsonResponse


class OnAppResponse(JsonResponse):
    """
    OnApp response class
    """

    def success(self):
        """
        Determine if our request was successful.

        The meaning of this can be arbitrary; did we receive OK status? Did
        the node get created? Were we authenticated?

        :rtype: ``bool``
        :return: ``True`` or ``False``
        """
        return self.status in [httplib.OK, httplib.CREATED, httplib.NO_CONTENT]


class OnAppConnection(ConnectionUserAndKey):
    """
    OnApp connection class
    """

    responseCls = OnAppResponse

    def add_default_headers(self, headers):
        """
        Add Basic Authentication header to all the requests.
        It injects the "Authorization: Basic Base64String===" header
        in each request

        :type  headers: ``dict``
        :param headers: Default input headers

        :rtype:         ``dict``
        :return:        Default input headers with the "Authorization" header.
        """
        b64string = b("%s:%s" % (self.user_id, self.key))
        encoded = b64encode(b64string).decode("utf-8")

        headers["Authorization"] = "Basic " + encoded
        return headers



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from libcloud.test import unittest

from libcloud.container.types import Provider
from libcloud.container.providers import get_driver

from libcloud.container.base import ContainerImage
from libcloud.container.base import Container

from libcloud.container.drivers.lxd import LXDStoragePool
from libcloud.container.drivers.lxd import LXDAPIException
from libcloud.container.drivers.lxd import LXDContainerDriver
from libcloud.container.drivers.lxd import LXDServerInfo

from libcloud.utils.py3 import httplib
from libcloud.test.secrets import CONTAINER_PARAMS_LXD
from libcloud.test.file_fixtures import ContainerFileFixtures
from libcloud.test import MockHttp


class LXDContainerDriverTestCase(unittest.TestCase):

    """
    Unit tests for LXDContainerDriver
    """

    def setUp(self):
        # Create a test driver for each version
        versions = ('linux_124',)
        self.drivers = []
        for version in versions:
            LXDContainerDriver.connectionCls.conn_class = \
                LXDMockHttp
            LXDMockHttp.type = None
            LXDMockHttp.use_param = 'a'
            driver = LXDContainerDriver(*CONTAINER_PARAMS_LXD)
            driver.connectionCls.conn_class = \
                LXDMockHttp
            driver.version = version
            self.drivers.append(driver)

    def test_ex_get_api_endpoints_trusted(self):
        for driver in self.drivers:
            api = driver.ex_get_api_endpoints()
            self.assertEqual(api[0], driver.version)

    def test_ex_get_server_configuration(self):
        for driver in self.drivers:
            server_config = driver.ex_get_server_configuration()
            self.assertIsInstance(server_config, LXDServerInfo)
            self.assertEqual(server_config.api_extensions, [])
            self.assertEqual(server_config.api_status, "stable")
            self.assertEqual(server_config.api_version, "linux_124")
            self.assertEqual(server_config.auth, "guest")
            self.assertEqual(server_config.public, False)

    def test_list_images(self):
        img_id = "54c8caac1f61901ed86c68f24af5f5d3672bdc62c71d04f06df3a59e95684473"
        for driver in self.drivers:
            images = driver.list_images()
            self.assertEqual(len(images), 1)
            self.assertIsInstance(images[0], ContainerImage)
            self.assertEqual(images[0].id, img_id)
            self.assertEqual(images[0].name, 'trusty')

    def test_list_containers(self):
        for driver in self.drivers:
            containers = driver.list_containers()
            self.assertEqual(len(containers), 2)
            self.assertIsInstance(containers[0], Container)
            self.assertIsInstance(containers[1], Container)
            self.assertEqual(containers[0].name, 'first_lxd_container')
            self.assertEqual(containers[1].name, 'second_lxd_container')

    def test_get_container(self):
        for driver in self.drivers:
            container = driver.get_container(id='second_lxd_container')
            self.assertIsInstance(container, Container)
            self.assertEqual(container.name, 'second_lxd_container')
            self.assertEqual(container.id, 'second_lxd_container')
            self.assertEqual(container.state, 'stopped')

    def test_start_container(self):
        for driver in self.drivers:
            container = driver.get_container(id='first_lxd_container')
            container.start()
            self.assertEqual(container.state, 'running')

    def test_stop_container(self):
        for driver in self.drivers:
            container = driver.get_container(id='second_lxd_container')
            container.stop()
            self.assertEqual(container.state, 'stopped')

    def test_restart_container(self):
        for driver in self.drivers:
            container = driver.get_container(id='second_lxd_container')
            container.restart()

    def test_delete_container(self):
        for driver in self.drivers:
            container = driver.get_container(id='second_lxd_container')
            container.destroy()

    def test_deploy_container(self):
        for driver in self.drivers:
            image = ContainerImage(id=None,
                                   name=None,
                                   path=None, version=None, driver=driver)
            container = driver.deploy_container(name='first_lxd_container',
                                                image=image,
                                                parameters='{"source":{"type":"image", '
                                                           '"fingerprint":"7ed08b435c92cd8a8a884c88e8722f2e7546a51e891982a90ea9c15619d7df9b"}}')
            self.assertIsInstance(container, Container)
            self.assertEqual(container.name, 'first_lxd_container')

    # test images
    def test_install_image_no_dict(self):
        with self.assertRaises(LXDAPIException) as exc:
            for driver in self.drivers:
                container = driver.install_image(path=None)
                self.assertEqual(str(exc), "Install an image for LXD requires specification of image_data")

    # test storage pools

    def test_list_storage_pools(self):
        for driver in self.drivers:
            pools = driver.ex_list_storage_pools()
            self.assertEqual(len(pools), 2)
            self.assertIsInstance(pools[0], LXDStoragePool)
            self.assertIsInstance(pools[1], LXDStoragePool)
            self.assertEqual(pools[0].name, 'pool1')
            self.assertEqual(pools[1].name, 'pool2')

    def test_get_storage_pool_no_metadata(self):
        with self.assertRaises(LXDAPIException) as exc:
            for driver in self.drivers:
                driver.ex_get_storage_pool(id='pool3')
                self.assertEqual(str(exc), 'Storage pool with name pool3 has no data')

    def test_delete_storage_pool(self):
        for driver in self.drivers:
            driver.ex_delete_storage_pool(id='pool1')

    def test_delete_storage_pool_fail(self):
        with self.assertRaises(LXDAPIException) as exc:
            for driver in self.drivers:
                driver.ex_delete_storage_pool(id='pool2')


class LXDMockHttp(MockHttp):
    fixtures = ContainerFileFixtures('lxd')
    version = None

    def _version(
        self, method, url, body, headers):
        if method == 'GET':
            body = self.fixtures.load('linux_124/version.json')
        else:
            raise AssertionError('Unsupported method')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def root(self, method, url, body, headers):
        json = self.fixtures.load('linux_124/endpoints_sucess.json')
        return (httplib.OK, json, {}, httplib.responses[httplib.OK])

    def _linux_124(
        self, method, url, body, headers):
        return (httplib.OK, self.fixtures.load('linux_124/version.json'), {}, httplib.responses[httplib.OK])

    def _linux_124_images(
        self, method, url, body, headers):
        return (httplib.OK, self.fixtures.load('linux_124/images.json'), {}, httplib.responses[httplib.OK])

    def _linux_124_images_54c8caac1f61901ed86c68f24af5f5d3672bdc62c71d04f06df3a59e95684473(self, method, url, body, headers):
        return (httplib.OK, self.fixtures.load('linux_124/image.json'), {}, httplib.responses[httplib.OK])

    def _linux_124_containers(
        self, method, url, body, headers):

        if method == 'GET':
            return (httplib.OK, self.fixtures.load('linux_124/containers.json'), {}, httplib.responses[httplib.OK])
        elif method == 'POST' or method == 'PUT':
            # we do a POST when we create a container
            # we will return a dummy background operation
            return (httplib.OK, self.fixtures.load('linux_124/background_op.json'), {}, httplib.responses[httplib.OK])

    def _linux_124_containers_first_lxd_container(self, method, url, body, headers):
        return (httplib.OK, self.fixtures.load('linux_124/first_lxd_container.json'), {}, httplib.responses[httplib.OK])

    def _linux_124_containers_second_lxd_container(
        self, method, url, body, headers):

        if method == 'PUT' or method == 'DELETE':
            json = self.fixtures.load('linux_124/background_op.json')
            return (httplib.OK, json, {}, httplib.responses[httplib.OK])
        elif method == 'GET':
            return (httplib.OK, self.fixtures.load('linux_124/second_lxd_container.json'), {}, httplib.responses[httplib.OK])

    def _linux_124_containers_first_lxd_container_state(self, method, url, body, headers):

        if method == 'PUT' or method == 'DELETE':
            json = self.fixtures.load('linux_124/background_op.json')
            return (httplib.OK, json, {}, httplib.responses[httplib.OK])
        elif method == 'GET':
            json = self.fixtures.load('linux_124/first_lxd_container.json')
            return (httplib.OK, json, {}, httplib.responses[httplib.OK])

    def _linux_124_containers_second_lxd_container_state(self, method, url, body, headers):

        if method == 'PUT' or method == 'DELETE':
            json = self.fixtures.load('linux_124/background_op.json')
            return (httplib.OK, json, {}, httplib.responses[httplib.OK])
        elif method == 'GET':
            json = self.fixtures.load('linux_124/second_lxd_container.json')
            return (httplib.OK, json, {}, httplib.responses[httplib.OK])

    def _linux_124_operations_1_wait(self, method, url, body, header):
        return (httplib.OK, self.fixtures.load("linux_124/operation_1_wait.json"), {}, httplib.responses[httplib.OK])

    def _linux_124_storage_pools(self, method, url, body, header):

        if method == 'GET':
            json = self.fixtures.load('linux_124/storage_pools.json')
            return (httplib.OK, json, {}, httplib.responses[httplib.OK])

    def _linux_124_storage_pools_pool1(self, method, url, body, header):

        if method == 'GET':
            json = self.fixtures.load('linux_124/storage_pool_1.json')
            return (httplib.OK, json, {}, httplib.responses[httplib.OK])
        elif method == 'DELETE':
            return (httplib.OK,
                    self.fixtures.load('linux_124/storage_pool_delete_sucess.json'),
                    {},
                    httplib.responses[httplib.OK])

    def _linux_124_storage_pools_pool2(self, method, url, body, header):

        if method == 'GET':
            json = self.fixtures.load('linux_124/storage_pool_2.json')
            return (httplib.OK, json, {}, httplib.responses[httplib.OK])
        elif method == 'DELETE':
            return (httplib.OK,
                    self.fixtures.load('linux_124/storage_pool_delete_fail.json'),
                    {},
                    httplib.responses[httplib.OK])

    def _linux_124_storage_pools_pool3(self, method, url, body, header):

        if method == 'GET':
            json = self.fixtures.load('linux_124/no_meta_pool.json')
            return (httplib.OK, json, {}, httplib.responses[httplib.OK])


    """
    def _vlinux_124_containers_create(
        self, method, url, body, headers):
        return (httplib.OK, self.fixtures.load('linux_124/create_container.json'), {}, httplib.responses[httplib.OK])

    def _vlinux_124_containers_a68c1872c74630522c7aa74b85558b06824c5e672cee334296c50fb209825303(
        self, method, url, body, headers):
        return (httplib.NO_CONTENT, '', {}, httplib.responses[httplib.OK])

    
    def _vlinux_124_containers_a68c1872c74630522c7aa74b85558b06824c5e672cee334296c50fb209825303_start(
        self, method, url, body, headers):
        return (httplib.NO_CONTENT, '', {}, httplib.responses[httplib.OK])

    def _vlinux_124_containers_a68c1872c74630522c7aa74b85558b06824c5e672cee334296c50fb209825303_restart(
        self, method, url, body, headers):
        return (httplib.NO_CONTENT, '', {}, httplib.responses[httplib.OK])

    def _vlinux_124_containers_a68c1872c74630522c7aa74b85558b06824c5e672cee334296c50fb209825303_rename(
        self, method, url, body, headers):
        return (httplib.NO_CONTENT, '', {}, httplib.responses[httplib.OK])

    def _vlinux_124_containers_a68c1872c74630522c7aa74b85558b06824c5e672cee334296c50fb209825303_stop(
        self, method, url, body, headers):
        return (httplib.NO_CONTENT, '', {}, httplib.responses[httplib.OK])

    def _vlinux_124_containers_a68c1872c74630522c7aa74b85558b06824c5e672cee334296c50fb209825303_json(
        self, method, url, body, headers):
        return (httplib.OK, self.fixtures.load('linux_124/container_a68.json'), {}, httplib.responses[httplib.OK])

    def _vlinux_124_containers_a68c1872c74630522c7aa74b85558b06824c5e672cee334296c50fb209825303_logs(
        self, method, url, body, headers):
        return (httplib.OK, self.fixtures.load('linux_124/logs.txt'), {'content-type': 'text/plain'},
                httplib.responses[httplib.OK])
    """

if __name__ == '__main__':
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.compute.providers import Provider
from libcloud.compute.base import Node, NodeImage, NodeSize
from libcloud.compute.drivers.cloudstack import CloudStackNodeDriver


class KTUCloudNodeDriver(CloudStackNodeDriver):
    """Driver for KTUCloud Compute platform."""

    EMPTY_DISKOFFERINGID = '0'
    type = Provider.KTUCLOUD
    name = 'KTUCloud'
    website = 'https://ucloudbiz.olleh.com/'

    def list_images(self, location=None):
        args = {
            'templatefilter': 'executable'
        }
        if location is not None:
            args['zoneid'] = location.id

        imgs = self._sync_request(command='listAvailableProductTypes',
                                  method='GET')
        images = []

        for img in imgs['producttypes']:
            images.append(
                NodeImage(
                    img['serviceofferingid'],
                    img['serviceofferingdesc'],
                    self,
                    {'hypervisor': '',
                     'format': '',
                     'os': img['templatedesc'],
                     'templateid': img['templateid'],
                     'zoneid': img['zoneid']}
                )
            )

        return images

    def list_sizes(self, location=None):
        szs = self._sync_request('listAvailableProductTypes')
        sizes = []
        for sz in szs['producttypes']:
            diskofferingid = sz.get('diskofferingid',
                                    self.EMPTY_DISKOFFERINGID)
            sizes.append(NodeSize(
                diskofferingid,
                sz['diskofferingdesc'],
                0, 0, 0, 0, self)
            )
        return sizes

    def create_node(self, name, size, image, location=None,
                    ex_usageplantype='hourly'):
        params = {'displayname': name,
                  'serviceofferingid': image.id,
                  'templateid': str(image.extra['templateid']),
                  'zoneid': str(image.extra['zoneid'])}

        if ex_usageplantype is None:
            params['usageplantype'] = 'hourly'
        else:
            params['usageplantype'] = ex_usageplantype

        if size.id != self.EMPTY_DISKOFFERINGID:
            params['diskofferingid'] = size.id

        result = self._async_request(
            command='deployVirtualMachine',
            params=params,
            method='GET')

        node = result['virtualmachine']

        return Node(
            id=node['id'],
            name=node['displayname'],
            state=self.NODE_STATE_MAP[node['state']],
            public_ips=[],
            private_ips=[],
            driver=self,
            extra={
                'zoneid': image.extra['zoneid'],
                'ip_addresses': [],
                'forwarding_rules': [],
            }
        )



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from libcloud.utils.py3 import httplib
from libcloud.utils.networking import is_private_subnet

from libcloud.common.onapp import OnAppConnection
from libcloud.compute.base import Node, NodeDriver, NodeImage, KeyPair
from libcloud.compute.providers import Provider


__all__ = [
    "OnAppNodeDriver"
]

"""
Define the extra dictionary for specific resources
"""
RESOURCE_EXTRA_ATTRIBUTES_MAP = {
    "node": {
        "add_to_marketplace": {
            "key_name": "add_to_marketplace",
            "transform_func": bool
        },
        "admin_note": {
            "key_name": "admin_note",
            "transform_func": str
        },
        "allow_resize_without_reboot": {
            "key_name": "allow_resize_without_reboot",
            "transform_func": bool
        },
        "allowed_hot_migrate": {
            "key_name": "allowed_hot_migrate",
            "transform_func": bool
        },
        "allowed_swap": {
            "key_name": "allowed_swap",
            "transform_func": bool
        },
        "booted": {
            "key_name": "booted",
            "transform_func": bool
        },
        "built": {
            "key_name": "built",
            "transform_func": bool
        },
        "cpu_priority": {
            "key_name": "cpu_priority",
            "transform_func": int
        },
        "cpu_shares": {
            "key_name": "cpu_shares",
            "transform_func": int
        },
        "cpu_sockets": {
            "key_name": "cpu_sockets",
            "transform_func": int
        },
        "cpu_threads": {
            "key_name": "cpu_threads",
            "transform_func": int
        },
        "cpu_units": {
            "key_name": "cpu_units",
            "transform_func": int
        },
        "cpus": {
            "key_name": "cpus",
            "transform_func": int
        },
        "created_at": {
            "key_name": "created_at",
            "transform_func": str
        },
        "customer_network_id": {
            "key_name": "customer_network_id",
            "transform_func": str
        },
        "deleted_at": {
            "key_name": "deleted_at",
            "transform_func": str
        },
        "edge_server_type": {
            "key_name": "edge_server_type",
            "transform_func": str
        },
        "enable_autoscale": {
            "key_name": "enable_autoscale",
            "transform_func": bool
        },
        "enable_monitis": {
            "key_name": "enable_monitis",
            "transform_func": bool
        },
        "firewall_notrack": {
            "key_name": "firewall_notrack",
            "transform_func": bool
        },
        "hostname": {
            "key_name": "hostname",
            "transform_func": str
        },
        "hypervisor_id": {
            "key_name": "hypervisor_id",
            "transform_func": int
        },
        "id": {
            "key_name": "id",
            "transform_func": int
        },
        "initial_root_password": {
            "key_name": "initial_root_password",
            "transform_func": str
        },
        "initial_root_password_encrypted": {
            "key_name": "initial_root_password_encrypted",
            "transform_func": bool
        },
        "local_remote_access_ip_address": {
            "key_name": "local_remote_access_ip_address",
            "transform_func": str
        },
        "local_remote_access_port": {
            "key_name": "local_remote_access_port",
            "transform_func": int
        },
        "locked": {
            "key_name": "locked",
            "transform_func": bool
        },
        "memory": {
            "key_name": "memory",
            "transform_func": int
        },
        "min_disk_size": {
            "key_name": "min_disk_size",
            "transform_func": int
        },
        "monthly_bandwidth_used": {
            "key_name": "monthly_bandwidth_used",
            "transform_func": int
        },
        "note": {
            "key_name": "note",
            "transform_func": str
        },
        "operating_system": {
            "key_name": "operating_system",
            "transform_func": str
        },
        "operating_system_distro": {
            "key_name": "operating_system_distro",
            "transform_func": str
        },
        "preferred_hvs": {
            "key_name": "preferred_hvs",
            "transform_func": list
        },
        "price_per_hour": {
            "key_name": "price_per_hour",
            "transform_func": float
        },
        "price_per_hour_powered_off": {
            "key_name": "price_per_hour_powered_off",
            "transform_func": float
        },
        "recovery_mode": {
            "key_name": "recovery_mode",
            "transform_func": bool
        },
        "remote_access_password": {
            "key_name": "remote_access_password",
            "transform_func": str
        },
        "service_password": {
            "key_name": "service_password",
            "transform_func": str
        },
        "state": {
            "key_name": "state",
            "transform_func": str
        },
        "storage_server_type": {
            "key_name": "storage_server_type",
            "transform_func": str
        },
        "strict_virtual_machine_id": {
            "key_name": "strict_virtual_machine_id",
            "transform_func": str
        },
        "support_incremental_backups": {
            "key_name": "support_incremental_backups",
            "transform_func": bool
        },
        "suspended": {
            "key_name": "suspended",
            "transform_func": bool
        },
        "template_id": {
            "key_name": "template_id",
            "transform_func": int
        },
        "template_label": {
            "key_name": "template_label",
            "transform_func": str
        },
        "total_disk_size": {
            "key_name": "total_disk_size",
            "transform_func": int
        },
        "updated_at": {
            "key_name": "updated_at",
            "transform_func": str
        },
        "user_id": {
            "key_name": "user_id",
            "transform_func": int
        },
        "vip": {
            "key_name": "vip",
            "transform_func": bool
        },
        "xen_id": {
            "key_name": "xen_id",
            "transform_func": int
        }
    }
}


class OnAppNodeDriver(NodeDriver):
    """
    Base OnApp node driver.
    """

    connectionCls = OnAppConnection
    type = Provider.ONAPP
    name = 'OnApp'
    website = 'http://onapp.com/'

    def create_node(self, name, ex_memory, ex_cpus, ex_cpu_shares,
                    ex_hostname, ex_template_id, ex_primary_disk_size,
                    ex_swap_disk_size, ex_required_virtual_machine_build=1,
                    ex_required_ip_address_assignment=1, **kwargs):
        """
        Add a VS

        :param  kwargs: All keyword arguments to create a VS
        :type   kwargs: ``dict``

        :rtype: :class:`OnAppNode`
        """
        server_params = dict(
            label=name,
            memory=ex_memory,
            cpus=ex_cpus,
            cpu_shares=ex_cpu_shares,
            hostname=ex_hostname,
            template_id=ex_template_id,
            primary_disk_size=ex_primary_disk_size,
            swap_disk_size=ex_swap_disk_size,
            required_virtual_machine_build=ex_required_virtual_machine_build,
            required_ip_address_assignment=ex_required_ip_address_assignment,
            rate_limit=kwargs.get("rate_limit")
        )

        server_params.update(OnAppNodeDriver._create_args_to_params(**kwargs))
        data = json.dumps({"virtual_machine": server_params})

        response = self.connection.request(
            "/virtual_machines.json",
            data=data,
            headers={
                "Content-type": "application/json"},
            method="POST")

        return self._to_node(response.object["virtual_machine"])

    def destroy_node(self,
                     node,
                     ex_convert_last_backup=0,
                     ex_destroy_all_backups=0):
        """
        Delete a VS

        :param node: OnApp node
        :type  node: :class: `OnAppNode`

        :param convert_last_backup: set 1 to convert the last VS's backup to
                                    template, otherwise set 0
        :type  convert_last_backup: ``int``

        :param destroy_all_backups: set 1 to destroy all existing backups of
                                    this VS, otherwise set 0
        :type  destroy_all_backups: ``int``
        """
        server_params = {
            "convert_last_backup": ex_convert_last_backup,
            "destroy_all_backups": ex_destroy_all_backups
        }
        action = "/virtual_machines/{identifier}.json".format(
            identifier=node.id)

        self.connection.request(action, params=server_params, method="DELETE")
        return True

    def list_nodes(self):
        """
        List all VS

        :rtype: ``list`` of :class:`OnAppNode`
        """
        response = self.connection.request("/virtual_machines.json")
        nodes = []
        for vm in response.object:
            nodes.append(self._to_node(vm["virtual_machine"]))
        return nodes

    def list_images(self):
        """
        List all images

        :rtype: ``list`` of :class:`NodeImage`
        """
        response = self.connection.request("/templates.json")
        templates = []
        for template in response.object:
            templates.append(self._to_image(template["image_template"]))
        return templates

    def list_key_pairs(self):
        """
        List all the available key pair objects.

        :rtype: ``list`` of :class:`.KeyPair` objects
        """
        user_id = self.connection.request('/profile.json').object['user']['id']
        response = self.connection.request('/users/%s/ssh_keys.json' % user_id)
        ssh_keys = []
        for ssh_key in response.object:
            ssh_keys.append(self._to_key_pair(ssh_key['ssh_key']))
        return ssh_keys

    def get_key_pair(self, name):
        """
        Retrieve a single key pair.

        :param name: ID of the key pair to retrieve.
        :type name: ``str``

        :rtype: :class:`.KeyPair` object
        """
        user_id = self.connection.request('/profile.json').object['user']['id']
        response = self.connection.request(
            '/users/%s/ssh_keys/%s.json' % (user_id, name))
        return self._to_key_pair(response.object['ssh_key'])

    def import_key_pair_from_string(self, name, key_material):
        """
        Import a new public key from string.

        :param name: Key pair name (unused).
        :type name: ``str``

        :param key_material: Public key material.
        :type key_material: ``str``

        :rtype: :class:`.KeyPair` object
        """
        data = json.dumps({'key': key_material})
        user_id = self.connection.request('/profile.json').object['user']['id']
        response = self.connection.request(
            '/users/%s/ssh_keys.json' % user_id,
            data=data,
            headers={
                "Content-type": "application/json"},
            method="POST")
        return self._to_key_pair(response.object['ssh_key'])

    def delete_key_pair(self, key):
        """
        Delete an existing key pair.

        :param key_pair: Key pair object.
        :type key_pair: :class:`.KeyPair`

        :return: True on success
        :rtype: ``bool``
        """
        key_id = key.name
        response = self.connection.request(
            '/settings/ssh_keys/%s.json' % key_id,
            method='DELETE')
        return response.status == httplib.NO_CONTENT

    #
    # Helper methods
    #

    def _to_key_pair(self, data):
        extra = {'created_at': data['created_at'],
                 'updated_at': data['updated_at']}
        return KeyPair(name=data['id'],
                       fingerprint=None,
                       public_key=data['key'],
                       private_key=None,
                       driver=self,
                       extra=extra)

    def _to_image(self, template):
        extra = {'distribution': template['operating_system_distro'],
                 'operating_system': template['operating_system'],
                 'operating_system_arch': template['operating_system_arch'],
                 'allow_resize_without_reboot':
                 template['allow_resize_without_reboot'],
                 'allowed_hot_migrate': template['allowed_hot_migrate'],
                 'allowed_swap': template['allowed_swap'],
                 'min_disk_size': template['min_disk_size'],
                 'min_memory_size': template['min_memory_size'],
                 'created_at': template['created_at']}
        return NodeImage(id=template['id'], name=template['label'],
                         driver=self, extra=extra)

    def _to_node(self, data):
        identifier = data["identifier"]
        name = data["label"]
        private_ips = []
        public_ips = []
        for ip in data["ip_addresses"]:
            address = ip["ip_address"]['address']
            if is_private_subnet(address):
                private_ips.append(address)
            else:
                public_ips.append(address)

        extra = OnAppNodeDriver._get_extra_dict(
            data, RESOURCE_EXTRA_ATTRIBUTES_MAP["node"]
        )
        return Node(identifier,
                    name,
                    extra['state'],
                    public_ips,
                    private_ips,
                    self,
                    extra=extra)

    @staticmethod
    def _get_extra_dict(response, mapping):
        """
        Extract attributes from the element based on rules provided in the
        mapping dictionary.

        :param   response: The JSON response to parse the values from.
        :type    response: ``dict``

        :param   mapping: Dictionary with the extra layout
        :type    mapping: ``dict``

        :rtype:  ``dict``
        """
        extra = {}
        for attribute, values in mapping.items():
            transform_func = values["transform_func"]
            value = response.get(values["key_name"])

            extra[attribute] = transform_func(value) if value else None
        return extra

    @staticmethod
    def _create_args_to_params(**kwargs):
        """
        Extract server params from keyword args to create a VS

        :param   kwargs: keyword args
        :return: ``dict``
        """
        params = [
            "ex_cpu_sockets",
            "ex_cpu_threads",
            "ex_enable_autoscale",
            "ex_data_store_group_primary_id",
            "ex_data_store_group_swap_id",
            "ex_hypervisor_group_id",
            "ex_hypervisor_id",
            "ex_initial_root_password",
            "ex_note",
            "ex_primary_disk_min_iops",
            "ex_primary_network_id",
            "ex_primary_network_group_id",
            "ex_recipe_ids",
            "ex_required_automatic_backup",
            "ex_required_virtual_machine_startup",
            "ex_required_virtual_machine_startup",
            "ex_selected_ip_address_id",
            "ex_swap_disk_min_iops",
            "ex_type_of_format",
            "ex_custom_recipe_variables",
            "ex_licensing_key",
            "ex_licensing_server_id",
            "ex_licensing_type",
        ]
        server_params = {}

        for p in params:
            value = kwargs.get(p)
            if value:
                server_params[p[3:]] = value
        return server_params



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from libcloud.utils.py3 import httplib
from libcloud.utils.py3 import ET
from libcloud.common.dimensiondata import DimensionDataAPIException
from libcloud.common.types import InvalidCredsError
from libcloud.backup.base import BackupTargetJob
from libcloud.backup.drivers.dimensiondata import DimensionDataBackupDriver as DimensionData
from libcloud.backup.drivers.dimensiondata import DEFAULT_BACKUP_PLAN

from libcloud.test import MockHttp, unittest
from libcloud.test.file_fixtures import BackupFileFixtures

from libcloud.test.secrets import DIMENSIONDATA_PARAMS


class DimensionData_v2_3_Tests(unittest.TestCase):

    def setUp(self):
        DimensionData.connectionCls.active_api_version = '2.3'
        DimensionData.connectionCls.conn_class = DimensionDataMockHttp
        DimensionDataMockHttp.type = None
        self.driver = DimensionData(*DIMENSIONDATA_PARAMS)

    def test_invalid_region(self):
        with self.assertRaises(ValueError):
            self.driver = DimensionData(*DIMENSIONDATA_PARAMS, region='blah')

    def test_invalid_creds(self):
        DimensionDataMockHttp.type = 'UNAUTHORIZED'
        with self.assertRaises(InvalidCredsError):
            self.driver.list_targets()

    def test_list_targets(self):
        targets = self.driver.list_targets()
        self.assertEqual(len(targets), 2)
        self.assertEqual(targets[0].id, '5579f3a7-4c32-4cf5-8a7e-b45c36a35c10')
        self.assertEqual(targets[0].address, 'e75ead52-692f-4314-8725-c8a4f4d13a87')
        self.assertEqual(targets[0].extra['servicePlan'], 'Enterprise')

    def test_create_target(self):
        target = self.driver.create_target(
            'name',
            'e75ead52-692f-4314-8725-c8a4f4d13a87',
            extra={'servicePlan': 'Enterprise'})
        self.assertEqual(target.id, 'ee7c4b64-f7af-4a4f-8384-be362273530f')
        self.assertEqual(target.address, 'e75ead52-692f-4314-8725-c8a4f4d13a87')
        self.assertEqual(target.extra['servicePlan'], 'Enterprise')

    def test_create_target_DEFAULT(self):
        DimensionDataMockHttp.type = 'DEFAULT'
        target = self.driver.create_target(
            'name',
            'e75ead52-692f-4314-8725-c8a4f4d13a87')
        self.assertEqual(target.id, 'ee7c4b64-f7af-4a4f-8384-be362273530f')
        self.assertEqual(target.address, 'e75ead52-692f-4314-8725-c8a4f4d13a87')

    def test_create_target_EXISTS(self):
        DimensionDataMockHttp.type = 'EXISTS'
        with self.assertRaises(DimensionDataAPIException) as context:
            self.driver.create_target(
                'name',
                'e75ead52-692f-4314-8725-c8a4f4d13a87',
                extra={'servicePlan': 'Enterprise'})
        self.assertEqual(context.exception.code, 'ERROR')
        self.assertEqual(context.exception.msg, 'Cloud backup for this server is already enabled or being enabled (state: NORMAL).')

    def test_update_target(self):
        target = self.driver.list_targets()[0]
        extra = {'servicePlan': 'Essentials'}
        new_target = self.driver.update_target(target, extra=extra)
        self.assertEqual(new_target.extra['servicePlan'], 'Essentials')

    def test_update_target_DEFAULT(self):
        DimensionDataMockHttp.type = 'DEFAULT'
        target = 'e75ead52-692f-4314-8725-c8a4f4d13a87'
        self.driver.update_target(target)

    def test_update_target_STR(self):
        target = 'e75ead52-692f-4314-8725-c8a4f4d13a87'
        extra = {'servicePlan': 'Essentials'}
        new_target = self.driver.update_target(target, extra=extra)
        self.assertEqual(new_target.extra['servicePlan'], 'Essentials')

    def test_delete_target(self):
        target = self.driver.list_targets()[0]
        self.assertTrue(self.driver.delete_target(target))

    def test_ex_add_client_to_target(self):
        target = self.driver.list_targets()[0]
        client = self.driver.ex_list_available_client_types(target)[0]
        storage_policy = self.driver.ex_list_available_storage_policies(target)[0]
        schedule_policy = self.driver.ex_list_available_schedule_policies(target)[0]
        self.assertTrue(
            self.driver.ex_add_client_to_target(target, client, storage_policy,
                                                schedule_policy, 'ON_FAILURE', 'nobody@example.com')
        )

    def test_ex_add_client_to_target_STR(self):
        self.assertTrue(
            self.driver.ex_add_client_to_target('e75ead52-692f-4314-8725-c8a4f4d13a87', 'FA.Linux', '14 Day Storage Policy',
                                                '12AM - 6AM', 'ON_FAILURE', 'nobody@example.com')
        )

    def test_ex_get_backup_details_for_target(self):
        target = self.driver.list_targets()[0]
        response = self.driver.ex_get_backup_details_for_target(target)
        self.assertEqual(response.service_plan, 'Enterprise')
        client = response.clients[0]
        self.assertEqual(client.id, '30b1ff76-c76d-4d7c-b39d-3b72be0384c8')
        self.assertEqual(client.type.type, 'FA.Linux')
        self.assertEqual(client.running_job.progress, 5)
        self.assertTrue(isinstance(client.running_job, BackupTargetJob))
        self.assertEqual(len(client.alert.notify_list), 2)
        self.assertTrue(isinstance(client.alert.notify_list, list))

    def test_ex_get_backup_details_for_target_NOBACKUP(self):
        target = self.driver.list_targets()[0].address
        DimensionDataMockHttp.type = 'NOBACKUP'
        response = self.driver.ex_get_backup_details_for_target(target)
        self.assertTrue(response is None)

    def test_ex_cancel_target_job(self):
        target = self.driver.list_targets()[0]
        response = self.driver.ex_get_backup_details_for_target(target)
        client = response.clients[0]
        self.assertTrue(isinstance(client.running_job, BackupTargetJob))
        success = client.running_job.cancel()
        self.assertTrue(success)

    def test_ex_cancel_target_job_with_extras(self):
        success = self.driver.cancel_target_job(
            None,
            ex_client='30b1ff76_c76d_4d7c_b39d_3b72be0384c8',
            ex_target='e75ead52_692f_4314_8725_c8a4f4d13a87'
        )
        self.assertTrue(success)

    def test_ex_cancel_target_job_FAIL(self):
        DimensionDataMockHttp.type = 'FAIL'
        with self.assertRaises(DimensionDataAPIException) as context:
            self.driver.cancel_target_job(
                None,
                ex_client='30b1ff76_c76d_4d7c_b39d_3b72be0384c8',
                ex_target='e75ead52_692f_4314_8725_c8a4f4d13a87'
            )
        self.assertEqual(context.exception.code, 'ERROR')

    """Test a backup info for a target that does not have a client"""
    def test_ex_get_backup_details_for_target_NO_CLIENT(self):
        DimensionDataMockHttp.type = 'NOCLIENT'
        response = self.driver.ex_get_backup_details_for_target('e75ead52-692f-4314-8725-c8a4f4d13a87')
        self.assertEqual(response.service_plan, 'Essentials')
        self.assertEqual(len(response.clients), 0)

    """Test a backup details that has a client, but no alerting or running jobs"""
    def test_ex_get_backup_details_for_target_NO_JOB_OR_ALERT(self):
        DimensionDataMockHttp.type = 'NOJOB'
        response = self.driver.ex_get_backup_details_for_target('e75ead52-692f-4314_8725-c8a4f4d13a87')
        self.assertEqual(response.service_plan, 'Enterprise')
        self.assertTrue(isinstance(response.clients, list))
        self.assertEqual(len(response.clients), 1)
        client = response.clients[0]
        self.assertEqual(client.id, '30b1ff76-c76d-4d7c-b39d-3b72be0384c8')
        self.assertEqual(client.type.type, 'FA.Linux')
        self.assertIsNone(client.running_job)
        self.assertIsNone(client.alert)

    """Test getting backup info for a server that doesn't exist"""
    def test_ex_get_backup_details_for_target_DISABLED(self):
        DimensionDataMockHttp.type = 'DISABLED'
        with self.assertRaises(DimensionDataAPIException) as context:
            self.driver.ex_get_backup_details_for_target('e75ead52-692f-4314-8725-c8a4f4d13a87')
        self.assertEqual(context.exception.code, 'ERROR')
        self.assertEqual(context.exception.msg, 'Server e75ead52-692f-4314-8725-c8a4f4d13a87 has not been provisioned for backup')

    def test_ex_list_available_client_types(self):
        target = self.driver.list_targets()[0]
        answer = self.driver.ex_list_available_client_types(target)
        self.assertEqual(len(answer), 1)
        self.assertEqual(answer[0].type, 'FA.Linux')
        self.assertEqual(answer[0].is_file_system, True)
        self.assertEqual(answer[0].description, 'Linux File system')

    def test_ex_list_available_storage_policies(self):
        target = self.driver.list_targets()[0]
        answer = self.driver.ex_list_available_storage_policies(target)
        self.assertEqual(len(answer), 1)
        self.assertEqual(answer[0].name,
                         '30 Day Storage Policy + Secondary Copy')
        self.assertEqual(answer[0].retention_period, 30)
        self.assertEqual(answer[0].secondary_location, 'Primary')

    def test_ex_list_available_schedule_policies(self):
        target = self.driver.list_targets()[0]
        answer = self.driver.ex_list_available_schedule_policies(target)
        self.assertEqual(len(answer), 1)
        self.assertEqual(answer[0].name, '12AM - 6AM')
        self.assertEqual(answer[0].description, 'Daily backup will start between 12AM - 6AM')

    def test_ex_remove_client_from_target(self):
        target = self.driver.list_targets()[0]
        client = self.driver.ex_get_backup_details_for_target('e75ead52-692f-4314-8725-c8a4f4d13a87').clients[0]
        self.assertTrue(self.driver.ex_remove_client_from_target(target, client))

    def test_ex_remove_client_from_target_STR(self):
        self.assertTrue(
            self.driver.ex_remove_client_from_target(
                'e75ead52-692f-4314-8725-c8a4f4d13a87',
                '30b1ff76-c76d-4d7c-b39d-3b72be0384c8'
            )
        )

    def test_ex_remove_client_from_target_FAIL(self):
        DimensionDataMockHttp.type = 'FAIL'
        with self.assertRaises(DimensionDataAPIException) as context:
            self.driver.ex_remove_client_from_target(
                'e75ead52-692f-4314-8725-c8a4f4d13a87',
                '30b1ff76-c76d-4d7c-b39d-3b72be0384c8'
            )
        self.assertEqual(context.exception.code, 'ERROR')
        self.assertTrue('Backup Client is currently performing another operation' in context.exception.msg)

    def test_priv_target_to_target_address(self):
        target = self.driver.list_targets()[0]
        self.assertEqual(
            self.driver._target_to_target_address(target),
            'e75ead52-692f-4314-8725-c8a4f4d13a87'
        )

    def test_priv_target_to_target_address_STR(self):
        self.assertEqual(
            self.driver._target_to_target_address('e75ead52-692f-4314-8725-c8a4f4d13a87'),
            'e75ead52-692f-4314-8725-c8a4f4d13a87'
        )

    def test_priv_target_to_target_address_TYPEERROR(self):
        with self.assertRaises(TypeError):
            self.driver._target_to_target_address([1, 2, 3])

    def test_priv_client_to_client_id(self):
        client = self.driver.ex_get_backup_details_for_target('e75ead52-692f-4314-8725-c8a4f4d13a87').clients[0]
        self.assertEqual(
            self.driver._client_to_client_id(client),
            '30b1ff76-c76d-4d7c-b39d-3b72be0384c8'
        )

    def test_priv_client_to_client_id_STR(self):
        self.assertEqual(
            self.driver._client_to_client_id('30b1ff76-c76d-4d7c-b39d-3b72be0384c8'),
            '30b1ff76-c76d-4d7c-b39d-3b72be0384c8'
        )

    def test_priv_client_to_client_id_TYPEERROR(self):
        with self.assertRaises(TypeError):
            self.driver._client_to_client_id([1, 2, 3])


class InvalidRequestError(Exception):
    def __init__(self, tag):
        super(InvalidRequestError, self).__init__("Invalid Request - %s" % tag)


class DimensionDataMockHttp(MockHttp):

    fixtures = BackupFileFixtures('dimensiondata')

    def _oec_0_9_myaccount_UNAUTHORIZED(self, method, url, body, headers):
        return (httplib.UNAUTHORIZED, "", {}, httplib.responses[httplib.UNAUTHORIZED])

    def _oec_0_9_myaccount(self, method, url, body, headers):
        body = self.fixtures.load('oec_0_9_myaccount.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_myaccount_EXISTS(self, method, url, body, headers):
        body = self.fixtures.load('oec_0_9_myaccount.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_myaccount_DEFAULT(self, method, url, body, headers):
        body = self.fixtures.load('oec_0_9_myaccount.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_myaccount_INPROGRESS(self, method, url, body, headers):
        body = self.fixtures.load('oec_0_9_myaccount.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_myaccount_FAIL(self, method, url, body, headers):
        body = self.fixtures.load('oec_0_9_myaccount.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_myaccount_NOCLIENT(self, method, url, body, headers):
        body = self.fixtures.load('oec_0_9_myaccount.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_myaccount_DISABLED(self, method, url, body, headers):
        body = self.fixtures.load('oec_0_9_myaccount.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_myaccount_NOJOB(self, method, url, body, headers):
        body = self.fixtures.load('oec_0_9_myaccount.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_3_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_server_e75ead52_692f_4314_8725_c8a4f4d13a87(self, method, url, body, headers):
        body = self.fixtures.load(
            'server_server_e75ead52_692f_4314_8725_c8a4f4d13a87.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_3_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_server_e75ead52_692f_4314_8725_c8a4f4d13a87_DEFAULT(self, method, url, body, headers):
        body = self.fixtures.load(
            'server_server_e75ead52_692f_4314_8725_c8a4f4d13a87_DEFAULT.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_3_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_server_e75ead52_692f_4314_8725_c8a4f4d13a87_NOCLIENT(self, method, url, body, headers):
        body = self.fixtures.load(
            'server_server_e75ead52_692f_4314_8725_c8a4f4d13a87_DEFAULT.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_3_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_server_e75ead52_692f_4314_8725_c8a4f4d13a87_NOJOB(self, method, url, body, headers):
        body = self.fixtures.load(
            'server_server_e75ead52_692f_4314_8725_c8a4f4d13a87_DEFAULT.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_3_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_server_e75ead52_692f_4314_8725_c8a4f4d13a87_DISABLED(self, method, url, body, headers):
        body = self.fixtures.load(
            'server_server_e75ead52_692f_4314_8725_c8a4f4d13a87_DEFAULT.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_3_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_server(self, method, url, body, headers):
        body = self.fixtures.load(
            'server_server.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_e75ead52_692f_4314_8725_c8a4f4d13a87_backup_client_type(self, method, url, body, headers):
        body = self.fixtures.load(
            '_backup_client_type.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_e75ead52_692f_4314_8725_c8a4f4d13a87_backup_client_storagePolicy(
            self, method, url, body, headers):
        body = self.fixtures.load(
            '_backup_client_storagePolicy.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_e75ead52_692f_4314_8725_c8a4f4d13a87_backup_client_schedulePolicy(
            self, method, url, body, headers):
        body = self.fixtures.load(
            '_backup_client_schedulePolicy.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_e75ead52_692f_4314_8725_c8a4f4d13a87_backup_client(
            self, method, url, body, headers):
        if method == 'POST':
            body = self.fixtures.load(
                '_backup_client_SUCCESS_PUT.xml')
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:
            raise ValueError("Unknown Method {0}".format(method))

    def _oec_0_9_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_e75ead52_692f_4314_8725_c8a4f4d13a87_backup_NOCLIENT(
            self, method, url, body, headers):
        # only gets here are implemented
        # If we get any other method something has gone wrong
        assert(method == 'GET')
        body = self.fixtures.load(
            '_backup_INFO_NOCLIENT.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_e75ead52_692f_4314_8725_c8a4f4d13a87_backup_DISABLED(
            self, method, url, body, headers):
        # only gets here are implemented
        # If we get any other method something has gone wrong
        assert(method == 'GET')
        body = self.fixtures.load(
            '_backup_INFO_DISABLED.xml')
        return (httplib.BAD_REQUEST, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_e75ead52_692f_4314_8725_c8a4f4d13a87_backup_NOJOB(
            self, method, url, body, headers):
        # only gets here are implemented
        # If we get any other method something has gone wrong
        assert(method == 'GET')
        body = self.fixtures.load(
            '_backup_INFO_NOJOB.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_e75ead52_692f_4314_8725_c8a4f4d13a87_backup_DEFAULT(
            self, method, url, body, headers):
        if method != 'POST':
            raise InvalidRequestError('Only POST is accepted for this test')
        request = ET.fromstring(body)
        service_plan = request.get('servicePlan')
        if service_plan != DEFAULT_BACKUP_PLAN:
            raise InvalidRequestError('The default plan %s should have been passed in.  Not %s' % (DEFAULT_BACKUP_PLAN, service_plan))
        body = self.fixtures.load(
            '_backup_ENABLE.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_e75ead52_692f_4314_8725_c8a4f4d13a87_backup(
            self, method, url, body, headers):
        if method == 'POST':
            body = self.fixtures.load(
                '_backup_ENABLE.xml')
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        elif method == 'GET':
            if url.endswith('disable'):
                body = self.fixtures.load(
                    '_backup_DISABLE.xml')
                return (httplib.OK, body, {}, httplib.responses[httplib.OK])
            body = self.fixtures.load(
                '_backup_INFO.xml')
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])

        else:
            raise ValueError("Unknown Method {0}".format(method))

    def _caas_2_3_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_server_e75ead52_692f_4314_8725_c8a4f4d13a87_NOBACKUP(
            self, method, url, body, headers):
        assert(method == 'GET')
        body = self.fixtures.load('server_server_NOBACKUP.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_e75ead52_692f_4314_8725_c8a4f4d13a87_backup_EXISTS(
            self, method, url, body, headers):
        # only POSTs are implemented
        # If we get any other method something has gone wrong
        assert(method == 'POST')
        body = self.fixtures.load(
            '_backup_EXISTS.xml')
        return (httplib.BAD_REQUEST, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_e75ead52_692f_4314_8725_c8a4f4d13a87_backup_modify(
            self, method, url, body, headers):
        request = ET.fromstring(body)
        service_plan = request.get('servicePlan')
        if service_plan != 'Essentials':
            raise InvalidRequestError("Expected Essentials backup plan in request")
        body = self.fixtures.load('_backup_modify.xml')

        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_e75ead52_692f_4314_8725_c8a4f4d13a87_backup_modify_DEFAULT(
            self, method, url, body, headers):
        request = ET.fromstring(body)
        service_plan = request.get('servicePlan')
        if service_plan != DEFAULT_BACKUP_PLAN:
            raise InvalidRequestError("Expected % backup plan in test" % DEFAULT_BACKUP_PLAN)
        body = self.fixtures.load('_backup_modify.xml')

        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_e75ead52_692f_4314_8725_c8a4f4d13a87_backup_client_30b1ff76_c76d_4d7c_b39d_3b72be0384c8(
            self, method, url, body, headers):
        if url.endswith('disable'):
            body = self.fixtures.load(
                ('_remove_backup_client.xml')
            )
        elif url.endswith('cancelJob'):
            body = self.fixtures.load(
                (''
                 '_backup_client_30b1ff76_c76d_4d7c_b39d_3b72be0384c8_cancelJob.xml')
            )
        else:
            raise ValueError("Unknown URL: %s" % url)
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_server_e75ead52_692f_4314_8725_c8a4f4d13a87_backup_client_30b1ff76_c76d_4d7c_b39d_3b72be0384c8_FAIL(
            self, method, url, body, headers):
        if url.endswith('disable'):
            body = self.fixtures.load(
                ('_remove_backup_client_FAIL.xml')
            )
        elif url.endswith('cancelJob'):
            body = self.fixtures.load(
                (''
                 '_backup_client_30b1ff76_c76d_4d7c_b39d_3b72be0384c8_cancelJob_FAIL.xml')
            )
        else:
            raise ValueError("Unknown URL: %s" % url)
        return (httplib.BAD_REQUEST, body, {}, httplib.responses[httplib.OK])


if __name__ == '__main__':
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tests for Google Container Engine Driver
"""

import sys
import unittest

from libcloud.utils.py3 import httplib
from libcloud.container.drivers.gke import GKEContainerDriver, API_VERSION
from libcloud.common.google import (GoogleBaseAuthConnection)
from libcloud.test.common.test_google import GoogleAuthMockHttp, GoogleTestCase

from libcloud.test import MockHttp
from libcloud.test.container import TestCaseMixin
from libcloud.test.file_fixtures import ContainerFileFixtures

from libcloud.test.secrets import GKE_PARAMS, GKE_KEYWORD_PARAMS


class GKEContainerDriverTestCase(GoogleTestCase, TestCaseMixin):
    """
    Google Compute Engine Test Class.
    """
    # Mock out a few specific calls that interact with the user, system or
    # environment.
    datacenter = 'us-central1-a'

    def setUp(self):
        GKEMockHttp.test = self
        GKEContainerDriver.connectionCls.conn_class = GKEMockHttp
        GoogleBaseAuthConnection.conn_class = GoogleAuthMockHttp
        GKEMockHttp.type = None
        kwargs = GKE_KEYWORD_PARAMS.copy()
        kwargs['auth_type'] = 'IA'
        kwargs['datacenter'] = self.datacenter
        self.driver = GKEContainerDriver(*GKE_PARAMS, **kwargs)

    def test_list_images_response(self):
        config = self.driver.list_clusters(ex_zone="us-central1-a")
        assert "clusters" in config
        assert config["clusters"][0]["zone"] == "us-central1-a"

    def test_server_config(self):
        config = self.driver.get_server_config()
        assert "validImageTypes" in config


class GKEMockHttp(MockHttp):
    fixtures = ContainerFileFixtures('gke')
    json_hdr = {'content-type': 'application/json; charset=UTF-8'}

    def _get_method_name(self, type, use_param, qs, path):
        api_path = '/%s' % API_VERSION
        project_path = '/projects/%s' % GKE_KEYWORD_PARAMS['project']
        path = path.replace(api_path, '')
        # This replace is separate, since there is a call with a different
        # project name
        path = path.replace(project_path, '')
        # The path to get project information is the base path, so use a fake
        # '/project' path instead
        if not path:
            path = '/project'
        method_name = super(GKEMockHttp, self)._get_method_name(
            type, use_param, qs, path)
        return method_name

    def _zones_us_central1_a_serverconfig(self, method, url, body, headers):
        body = self.fixtures.load(
            'zones_us-central1-a_instance_serverconfig.json')
        return (httplib.OK, body, self.json_hdr, httplib.responses[httplib.OK])

    def _zones_us_central1_a_clusters(self, method, url, body, headers):
        body = self.fixtures.load(
            'zones_us-central1-a_list.json')
        return (httplib.OK, body, self.json_hdr, httplib.responses[httplib.OK])


if __name__ == '__main__':
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Dummy Driver

@note: This driver is out of date
"""
import uuid
import socket
import struct

from libcloud.common.base import ConnectionKey
from libcloud.compute.base import NodeImage, NodeSize, Node
from libcloud.compute.base import NodeDriver, NodeLocation
from libcloud.compute.base import KeyPair
from libcloud.compute.types import Provider, NodeState


class DummyConnection(ConnectionKey):
    """
    Dummy connection class
    """

    def connect(self, host=None, port=None):
        pass


class DummyNodeDriver(NodeDriver):
    """
    Dummy node driver

    This is a fake driver which appears to always create or destroy
    nodes successfully.

    >>> from libcloud.compute.drivers.dummy import DummyNodeDriver
    >>> driver = DummyNodeDriver(0)
    >>> node=driver.create_node()
    >>> node.public_ips[0]
    '127.0.0.3'
    >>> node.name
    'dummy-3'

    If the credentials you give convert to an integer then the next
    node to be created will be one higher.

    Each time you create a node you will get a different IP address.

    >>> driver = DummyNodeDriver(22)
    >>> node=driver.create_node()
    >>> node.name
    'dummy-23'

    """

    name = "Dummy Node Provider"
    website = 'http://example.com'
    type = Provider.DUMMY

    def __init__(self, creds):
        """
        :param  creds: Credentials
        :type   creds: ``str``

        :rtype: ``None``
        """
        self.creds = creds
        try:
            num = int(creds)
        except ValueError:
            num = None
        if num:
            self.nl = []
            startip = _ip_to_int('127.0.0.1')
            for i in range(num):
                ip = _int_to_ip(startip + i)
                self.nl.append(
                    Node(id=i,
                         name='dummy-%d' % (i),
                         state=NodeState.RUNNING,
                         public_ips=[ip],
                         private_ips=[],
                         driver=self,
                         extra={'foo': 'bar'})
                )
        else:
            self.nl = [
                Node(id=1,
                     name='dummy-1',
                     state=NodeState.RUNNING,
                     public_ips=['127.0.0.1'],
                     private_ips=[],
                     driver=self,
                     extra={'foo': 'bar'}),
                Node(id=2,
                     name='dummy-2',
                     state=NodeState.RUNNING,
                     public_ips=['127.0.0.1'],
                     private_ips=[],
                     driver=self,
                     extra={'foo': 'bar'}),
            ]
        self.connection = DummyConnection(self.creds)

    def get_uuid(self, unique_field=None):
        """

        :param  unique_field: Unique field
        :type   unique_field: ``bool``
        :rtype: :class:`UUID`
        """
        return str(uuid.uuid4())

    def list_nodes(self):
        """
        List the nodes known to a particular driver;
        There are two default nodes created at the beginning

        >>> from libcloud.compute.drivers.dummy import DummyNodeDriver
        >>> driver = DummyNodeDriver(0)
        >>> node_list=driver.list_nodes()
        >>> sorted([node.name for node in node_list ])
        ['dummy-1', 'dummy-2']

        each item in the list returned is a node object from which you
        can carry out any node actions you wish

        >>> node_list[0].reboot()
        True

        As more nodes are added, list_nodes will return them

        >>> node=driver.create_node()
        >>> node.size.id
        's1'
        >>> node.image.id
        'i2'
        >>> sorted([n.name for n in driver.list_nodes()])
        ['dummy-1', 'dummy-2', 'dummy-3']

        @inherits: :class:`NodeDriver.list_nodes`
        """
        return self.nl

    def reboot_node(self, node):
        """
        Sets the node state to rebooting; in this dummy driver always
        returns True as if the reboot had been successful.

        >>> from libcloud.compute.drivers.dummy import DummyNodeDriver
        >>> driver = DummyNodeDriver(0)
        >>> node=driver.create_node()
        >>> from libcloud.compute.types import NodeState
        >>> node.state == NodeState.RUNNING
        True
        >>> node.state == NodeState.REBOOTING
        False
        >>> driver.reboot_node(node)
        True
        >>> node.state == NodeState.REBOOTING
        True

        Please note, dummy nodes never recover from the reboot.

        @inherits: :class:`NodeDriver.reboot_node`
        """

        node.state = NodeState.REBOOTING
        return True

    def destroy_node(self, node):
        """
        Sets the node state to terminated and removes it from the node list

        >>> from libcloud.compute.drivers.dummy import DummyNodeDriver
        >>> driver = DummyNodeDriver(0)
        >>> from libcloud.compute.types import NodeState
        >>> node = [node for node in driver.list_nodes() if
        ...         node.name == 'dummy-1'][0]
        >>> node.state == NodeState.RUNNING
        True
        >>> driver.destroy_node(node)
        True
        >>> node.state == NodeState.RUNNING
        False
        >>> [n for n in driver.list_nodes() if n.name == 'dummy-1']
        []

        @inherits: :class:`NodeDriver.destroy_node`
        """

        node.state = NodeState.TERMINATED
        self.nl.remove(node)
        return True

    def list_images(self, location=None):
        """
        Returns a list of images as a cloud provider might have

        >>> from libcloud.compute.drivers.dummy import DummyNodeDriver
        >>> driver = DummyNodeDriver(0)
        >>> sorted([image.name for image in driver.list_images()])
        ['Slackware 4', 'Ubuntu 9.04', 'Ubuntu 9.10']

        @inherits: :class:`NodeDriver.list_images`
        """
        return [
            NodeImage(id=1, name="Ubuntu 9.10", driver=self),
            NodeImage(id=2, name="Ubuntu 9.04", driver=self),
            NodeImage(id=3, name="Slackware 4", driver=self),
        ]

    def list_sizes(self, location=None):
        """
        Returns a list of node sizes as a cloud provider might have

        >>> from libcloud.compute.drivers.dummy import DummyNodeDriver
        >>> driver = DummyNodeDriver(0)
        >>> sorted([size.ram for size in driver.list_sizes()])
        [128, 512, 4096, 8192]

        @inherits: :class:`NodeDriver.list_images`
        """

        return [
            NodeSize(id=1,
                     name="Small",
                     ram=128,
                     disk=4,
                     bandwidth=500,
                     price=4,
                     driver=self),
            NodeSize(id=2,
                     name="Medium",
                     ram=512,
                     disk=16,
                     bandwidth=1500,
                     price=8,
                     driver=self),
            NodeSize(id=3,
                     name="Big",
                     ram=4096,
                     disk=32,
                     bandwidth=2500,
                     price=32,
                     driver=self),
            NodeSize(id=4,
                     name="XXL Big",
                     ram=4096 * 2,
                     disk=32 * 4,
                     bandwidth=2500 * 3,
                     price=32 * 2,
                     driver=self),
        ]

    def list_locations(self):
        """
        Returns a list of locations of nodes

        >>> from libcloud.compute.drivers.dummy import DummyNodeDriver
        >>> driver = DummyNodeDriver(0)
        >>> sorted([loc.name + " in " + loc.country for loc in
        ...         driver.list_locations()])
        ['Island Datacenter in FJ', 'London Loft in GB', "Paul's Room in US"]

        @inherits: :class:`NodeDriver.list_locations`
        """
        return [
            NodeLocation(id=1,
                         name="Paul's Room",
                         country='US',
                         driver=self),
            NodeLocation(id=2,
                         name="London Loft",
                         country='GB',
                         driver=self),
            NodeLocation(id=3,
                         name="Island Datacenter",
                         country='FJ',
                         driver=self),
        ]

    def create_node(self, name, size, image):
        """
        Creates a dummy node; the node id is equal to the number of
        nodes in the node list

        >>> from libcloud.compute.drivers.dummy import DummyNodeDriver
        >>> driver = DummyNodeDriver(0)
        >>> sorted([node.name for node in driver.list_nodes()])
        ['dummy-1', 'dummy-2']
        >>> nodeA = driver.create_node()
        >>> sorted([node.name for node in driver.list_nodes()])
        ['dummy-1', 'dummy-2', 'dummy-3']
        >>> driver.create_node().name
        'dummy-4'
        >>> driver.destroy_node(nodeA)
        True
        >>> sorted([node.name for node in driver.list_nodes()])
        ['dummy-1', 'dummy-2', 'dummy-4']

        @inherits: :class:`NodeDriver.create_node`
        """
        num = len(self.nl) + 1
        n = Node(id=num,
                 name='dummy-%d' % (num),
                 state=NodeState.RUNNING,
                 public_ips=['127.0.0.%d' % (num)],
                 private_ips=[],
                 driver=self,
                 size=NodeSize(id='s1', name='foo', ram=2048,
                               disk=160, bandwidth=None, price=0.0,
                               driver=self),
                 image=NodeImage(id='i2', name='image', driver=self),
                 extra={'foo': 'bar'})
        self.nl.append(n)
        return n

    def import_key_pair_from_string(self, name, key_material):
        key_pair = KeyPair(name=name,
                           public_key=key_material,
                           fingerprint='fingerprint',
                           private_key='private_key',
                           driver=self)
        return key_pair


def _ip_to_int(ip):
    return socket.htonl(struct.unpack('I', socket.inet_aton(ip))[0])


def _int_to_ip(ip):
    return socket.inet_ntoa(struct.pack('I', socket.ntohl(ip)))


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from libcloud.test import unittest

from libcloud.utils.py3 import httplib
from libcloud.dns.types import RecordDoesNotExistError
from libcloud.dns.types import RecordType
from libcloud.dns.types import ZoneDoesNotExistError
from libcloud.dns.drivers.softlayer import SoftLayerDNSDriver
from libcloud.test import MockHttp
from libcloud.test.file_fixtures import DNSFileFixtures
from libcloud.test.secrets import SOFTLAYER_PARAMS
from libcloud.utils.py3 import xmlrpclib


class SoftLayerTests(unittest.TestCase):

    def setUp(self):
        SoftLayerDNSDriver.connectionCls.conn_class = SoftLayerDNSMockHttp
        SoftLayerDNSMockHttp.type = None
        self.driver = SoftLayerDNSDriver(*SOFTLAYER_PARAMS)

    def test_create_zone(self):
        zone = self.driver.create_zone(domain='bar.com')
        self.assertEqual(zone.id, '123')
        self.assertEqual(zone.domain, 'bar.com')

    def test_list_zones(self):
        zones = self.driver.list_zones()
        self.assertEqual(len(zones), 1)

        zone = zones[0]
        self.assertEqual(zone.id, '123')
        self.assertEqual(zone.type, 'master')
        self.assertEqual(zone.domain, 'bar.com')

    def test_get_zone(self):
        zone = self.driver.get_zone(zone_id='123')
        self.assertEqual(zone.id, '123')
        self.assertEqual(zone.type, 'master')
        self.assertEqual(zone.domain, 'bar.com')

    def test_get_zone_does_not_exist(self):
        SoftLayerDNSMockHttp.type = 'ZONE_DOES_NOT_EXIST'

        with self.assertRaises(ZoneDoesNotExistError):
            self.driver.get_zone(zone_id='333')

    def test_delete_zone(self):
        zone = self.driver.list_zones()[0]
        status = self.driver.delete_zone(zone=zone)
        self.assertTrue(status)

    def test_delete_zone_does_not_exist(self):
        zone = self.driver.list_zones()[0]

        SoftLayerDNSMockHttp.type = 'ZONE_DOES_NOT_EXIST'

        with self.assertRaises(ZoneDoesNotExistError):
            self.driver.delete_zone(zone=zone)

    def test_list_records(self):
        zone = self.driver.list_zones()[0]

        records = zone.list_records()

        self.assertEqual(records[0].id, '50772366')
        self.assertEqual(records[0].type, RecordType.SOA)
        self.assertEqual(records[0].data, 'ns1.softlayer.com.')
        self.assertEqual(
            records[0].extra,
            {
                'mxPriority': '',
                'expire': 604800,
                'retry': 300,
                'refresh': 3600,
                'ttl': 86400
            }
        )
        self.assertEqual(records[1].id, '50772367')
        self.assertEqual(records[1].type, RecordType.NS)
        self.assertEqual(records[1].data, 'ns1.softlayer.com.')

        self.assertEqual(records[2].id, '50772368')
        self.assertEqual(records[2].type, RecordType.NS)
        self.assertEqual(records[2].data, 'ns2.softlayer.com.')

        self.assertEqual(records[3].id, '50772365')
        self.assertEqual(records[3].type, RecordType.A)
        self.assertEqual(records[3].data, '127.0.0.1')

    def test_list_record_types(self):
        record_types = self.driver.list_record_types()
        self.assertEqual(len(record_types), 10)
        self.assertTrue(RecordType.A in record_types)

    def test_get_record(self):
        record = self.driver.get_record(zone_id='123', record_id='50772366')

        self.assertEqual(record.id, '50772366')
        self.assertEqual(record.type, RecordType.SOA)
        self.assertEqual(record.data, 'ns1.softlayer.com.')

    def test_get_record_record_does_not_exist(self):
        SoftLayerDNSMockHttp.type = 'RECORD_DOES_NOT_EXIST'

        with self.assertRaises(RecordDoesNotExistError):
            self.driver.get_record(zone_id='123',
                                   record_id='1')

    def test_delete_record(self):
        record = self.driver.get_record(zone_id='123', record_id='50772366')
        status = self.driver.delete_record(record=record)
        self.assertTrue(status)

    def test_delete_record_does_not_exist(self):
        record = self.driver.get_record(zone_id='123', record_id='50772366')

        SoftLayerDNSMockHttp.type = 'RECORD_DOES_NOT_EXIST'

        with self.assertRaises(RecordDoesNotExistError):
            self.driver.delete_record(record=record)

    def test_create_record(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.create_record(
            name='www', zone=zone,
            type=RecordType.A, data='127.0.0.1',
            extra={'ttl': 30}
        )

        self.assertEqual(record.id, '50772870')
        self.assertEqual(record.name, 'www')
        self.assertEqual(record.zone, zone)
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, '127.0.0.1')

    def test_update_record(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone=zone)[1]

        SoftLayerDNSMockHttp.type = 'CHANGED'
        params = {
            'record': record,
            'name': 'www',
            'type': RecordType.A,
            'data': '1.1.1.1',
            'extra': {'ttl': 30}}
        updated_record = self.driver.update_record(**params)

        self.assertEqual(record.data, 'ns1.softlayer.com.')

        self.assertEqual(updated_record.id, '123')
        self.assertEqual(updated_record.name, 'www')
        self.assertEqual(updated_record.zone, record.zone)
        self.assertEqual(updated_record.type, RecordType.A)
        self.assertEqual(updated_record.data, '1.1.1.1')


class SoftLayerDNSMockHttp(MockHttp):
    fixtures = DNSFileFixtures('softlayer')

    def _get_method_name(self, type, use_param, qs, path):
        return "_xmlrpc"

    def _xmlrpc(self, method, url, body, headers):
        params, meth_name = xmlrpclib.loads(body)
        url = url.replace("/", "_")
        meth_name = "%s_%s" % (url, meth_name)
        return getattr(self, meth_name)(method, url, body, headers)

    def _xmlrpc_v3_SoftLayer_Dns_Domain_createObject(
            self, method, url, body, headers):
        body = self.fixtures.load(
            'v3_SoftLayer_Dns_Domain_createObject.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc_v3_SoftLayer_Dns_Domain_getByDomainName(
            self, method, url, body, headers):
        body = self.fixtures.load(
            'v3_SoftLayer_Dns_Domain_getByDomainName.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc_v3_SoftLayer_Dns_Domain_getObject(
            self, method, url, body, headers):
        fixture = {
            None: 'v3_SoftLayer_Dns_Domain_getObject.xml',
            'ZONE_DOES_NOT_EXIST': 'not_found.xml',
        }[self.type]
        body = self.fixtures.load(fixture)
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc_v3_SoftLayer_Dns_Domain_deleteObject(
            self, method, url, body, headers):
        fixture = {
            None: 'v3_SoftLayer_Dns_Domain_deleteObject.xml',
            'ZONE_DOES_NOT_EXIST': 'not_found.xml',
        }[self.type]
        body = self.fixtures.load(fixture)
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc_v3_SoftLayer_Dns_Domain_getResourceRecords(
            self, method, url, body, headers):
        body = self.fixtures.load(
            'v3_SoftLayer_Dns_Domain_getResourceRecords.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc_v3_SoftLayer_Dns_Domain_ResourceRecord_getObject(
            self, method, url, body, headers):
        fixture = {
            None: 'v3_SoftLayer_Dns_Domain_ResourceRecord_getObject.xml',
            'RECORD_DOES_NOT_EXIST': 'not_found.xml',
            'CHANGED': 'v3_SoftLayer_Dns_Domain_ResourceRecord_getObject_changed.xml',
        }[self.type]
        body = self.fixtures.load(fixture)
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc_v3_SoftLayer_Dns_Domain_ResourceRecord_deleteObject(
            self, method, url, body, headers):
        fixture = {
            None: 'v3_SoftLayer_Dns_Domain_ResourceRecord_deleteObject.xml',
            'RECORD_DOES_NOT_EXIST': 'not_found.xml',
        }[self.type]
        body = self.fixtures.load(fixture)
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc_v3_SoftLayer_Dns_Domain_ResourceRecord_createObject(
            self, method, url, body, headers):
        body = self.fixtures.load(
            'v3_SoftLayer_Dns_Domain_ResourceRecord_createObject.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc_v3_SoftLayer_Dns_Domain_ResourceRecord_editObject(
            self, method, url, body, headers):
        body = self.fixtures.load(
            'v3_SoftLayer_Dns_Domain_ResourceRecord_editObject.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

if __name__ == '__main__':
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Equinix Metal Driver
"""

try:  # Try to use asyncio to perform requests in parallel across projects
    import asyncio
except ImportError:  # If not available will do things serially
    asyncio = None

import datetime
import json

from libcloud.utils.py3 import httplib

from libcloud.common.base import ConnectionKey, JsonResponse
from libcloud.compute.types import Provider, NodeState, InvalidCredsError
from libcloud.compute.base import NodeDriver, Node
from libcloud.compute.base import NodeImage, NodeSize, NodeLocation
from libcloud.compute.base import KeyPair
from libcloud.compute.base import StorageVolume, VolumeSnapshot

EQUINIXMETAL_ENDPOINT = "api.equinix.com"

# True to use async io if available (aka running under Python 3)
USE_ASYNC_IO_IF_AVAILABLE = True


def use_asyncio():
    return asyncio is not None and USE_ASYNC_IO_IF_AVAILABLE


class EquinixMetalResponse(JsonResponse):
    valid_response_codes = [httplib.OK, httplib.ACCEPTED, httplib.CREATED,
                            httplib.NO_CONTENT]

    def parse_error(self):
        if self.status == httplib.UNAUTHORIZED:
            body = self.parse_body()
            raise InvalidCredsError(body.get('error'))
        else:
            body = self.parse_body()
            if 'message' in body:
                error = '%s (code: %s)' % (body.get('message'), self.status)
            elif 'errors' in body:
                error = body.get('errors')
            else:
                error = body
            raise Exception(error)

    def success(self):
        return self.status in self.valid_response_codes


class EquinixMetalConnection(ConnectionKey):
    """
    Connection class for the Equinix Metal driver.
    """

    host = EQUINIXMETAL_ENDPOINT
    responseCls = EquinixMetalResponse

    def add_default_headers(self, headers):
        """
        Add headers that are necessary for every request
        """
        headers['Content-Type'] = 'application/json'
        headers['X-Auth-Token'] = self.key
        headers['X-Consumer-Token'] = \
            'kcrhMn7hwG8Ceo2hAhGFa2qpxLBvVHxEjS9ue8iqmsNkeeB2iQgMq4dNc1893pYu'
        return headers


class EquinixMetalNodeDriver(NodeDriver):
    """
    Equinix Metal NodeDriver
    """

    connectionCls = EquinixMetalConnection
    type = Provider.EQUINIXMETAL
    name = 'EquinixMetal'
    website = 'https://metal.equinix.com/'

    NODE_STATE_MAP = {'queued': NodeState.PENDING,
                      'provisioning': NodeState.PENDING,
                      'rebuilding': NodeState.PENDING,
                      'powering_on': NodeState.REBOOTING,
                      'powering_off': NodeState.REBOOTING,
                      'rebooting': NodeState.REBOOTING,
                      'inactive': NodeState.STOPPED,
                      'deleted': NodeState.TERMINATED,
                      'deprovisioning': NodeState.TERMINATED,
                      'failed': NodeState.ERROR,
                      'active': NodeState.RUNNING}

    def __init__(self, key, project=None):
        """
        Initialize a NodeDriver for Equinix Metal using the API token
        and optionally the project (name or id).

        If project name is specified we validate it lazily and populate
        self.project_id during the first access of self.projects variable
        """
        super(EquinixMetalNodeDriver, self).__init__(key=key)

        self.project_name = project
        self.project_id = None

        # Lazily populated on first access to self.project
        self._project = project

        # Variable which indicates if self._projects has been populated yet and
        # has been called self._project validated
        self._projects_populated = False
        self._projects = None

    @property
    def projects(self):
        """
        Lazily retrieve projects and set self.project_id variable on initial
        access to self.projects variable.
        """
        if not self._projects_populated:
            # NOTE: Each EquinixMetal account needs at least one project,
            # but to be on the safe side and avoid infinite loop
            # in case there are no projects on the account, we don't use
            # a more robust way to determine
            # if project list has been populated yet
            self._projects = self.ex_list_projects()
            self._projects_populated = True

            # If project name is specified, verify it's valid and populate
            # self.project_id
            if self._project:
                for project_obj in self._projects:
                    if self._project in [project_obj.name, project_obj.id]:
                        self.project_id = project_obj.id
                        break

                if not self.project_id:
                    # Invalid project name
                    self.project_name = None

        return self._projects

    def ex_list_projects(self):
        projects = []
        data = self.connection.request('/metal/v1/projects').object
        projects = data.get('projects')
        if projects:
            projects = [Project(project) for project in projects]
        return projects

    def list_nodes(self, ex_project_id=None):
        if ex_project_id:
            return self.ex_list_nodes_for_project(ex_project_id=ex_project_id)

        # if project has been specified during driver initialization, then
        # return nodes for this project only
        if self.project_id:
            return self.ex_list_nodes_for_project(
                ex_project_id=self.project_id)

        # In case of Python2 perform requests serially
        if not use_asyncio():
            nodes = []
            for project in self.projects:
                nodes.extend(
                    self.ex_list_nodes_for_project(ex_project_id=project.id)
                )
            return nodes
        # In case of Python3 use asyncio to perform requests in parallel
        return self.list_resources_async('nodes')

    def list_resources_async(self, resource_type):
        # The _list_nodes function is defined dynamically using exec in
        # order to prevent a SyntaxError in Python2 due to "yield from".
        # This cruft can be removed once Python2 support is no longer
        # required.
        assert resource_type in ['nodes', 'volumes']
        glob = globals()
        loc = locals()
        exec("""
import asyncio
@asyncio.coroutine
def _list_async(driver):
    projects = [project.id for project in driver.projects]
    loop = asyncio.get_event_loop()
    futures = [
        loop.run_in_executor(None, driver.ex_list_%s_for_project, p)
        for p in projects
    ]
    retval = []
    for future in futures:
        result = yield from future
        retval.extend(result)
    return retval""" % resource_type, glob, loc)
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            asyncio.set_event_loop(asyncio.new_event_loop())
            loop = asyncio.get_event_loop()

        return loop.run_until_complete(loc['_list_async'](loc['self']))

    def ex_list_nodes_for_project(self, ex_project_id, include='plan', page=1,
                                  per_page=1000):
        params = {
            'include': include,
            'page': page,
            'per_page': per_page
        }
        data = self.connection.request(
            '/metal/v1/projects/%s/devices' % (ex_project_id),
            params=params).object['devices']
        return list(map(self._to_node, data))

    def list_locations(self):
        data = self.connection.request('/metal/v1/facilities')\
            .object['facilities']
        return list(map(self._to_location, data))

    def list_images(self):
        data = self.connection.request('/metal/v1/operating-systems')\
            .object['operating_systems']
        return list(map(self._to_image, data))

    def list_sizes(self, ex_project_id=None):
        project_id = ex_project_id or self.project_id or (
            len(self.projects) and self.projects[0].id)
        if project_id:
            data = self.connection.request('/metal/v1/projects/%s/plans' %
                                           project_id).object['plans']
        else:  # This only works with personal tokens
            data = self.connection.request('/metal/v1/plans').object['plans']
        return [self._to_size(size) for size in data if
                size.get('line') == 'baremetal']

    def create_node(self, name, size, image, location,
                    ex_project_id=None, ip_addresses=[], cloud_init=None,
                    disk=None, disk_size=0, **kwargs):
        """
        Create a node.

        :return: The newly created node.
        :rtype: :class:`Node`
        """
        # if project has been specified on initialization of driver, then
        # create on this project

        if self.project_id:
            ex_project_id = self.project_id
        else:
            if not ex_project_id:
                raise Exception('ex_project_id needs to be specified')

        facility = location.extra['code']
        params = {'hostname': name, 'plan': size.id,
                  'operating_system': image.id, 'facility': facility,
                  'include': 'plan', 'billing_cycle': 'hourly'}
        if ip_addresses:
            params['ip_addresses'] = ip_addresses
        params.update(kwargs)
        if cloud_init:
            params["userdata"] = cloud_init
        data = self.connection.request('/metal/v1/projects/%s/devices' %
                                       (ex_project_id),
                                       data=json.dumps(params), method='POST')

        status = data.object.get('status', 'OK')
        if status == 'ERROR':
            message = data.object.get('message', None)
            error_message = data.object.get('error_message', message)
            raise ValueError('Failed to create node: %s' % (error_message))
        node = self._to_node(data=data.object)
        if disk:
            self.attach_volume(node, disk)
        if disk_size:
            volume = self.create_volume(size=disk_size,
                                        location=location)
            self.attach_volume(node, volume)
        return node

    def reboot_node(self, node):
        params = {'type': 'reboot'}
        res = self.connection.request('/metal/v1/devices/%s/actions'
                                      % (node.id),
                                      params=params, method='POST')
        return res.status == httplib.OK

    def start_node(self, node):
        params = {'type': 'power_on'}
        res = self.connection.request('/metal/v1/devices/%s/actions'
                                      % (node.id),
                                      params=params, method='POST')
        return res.status == httplib.OK

    def stop_node(self, node):
        params = {'type': 'power_off'}
        res = self.connection.request('/metal/v1/devices/%s/actions'
                                      % (node.id),
                                      params=params, method='POST')
        return res.status == httplib.OK

    def destroy_node(self, node):
        res = self.connection.request('/metal/v1/devices/%s' % (node.id),
                                      method='DELETE')
        return res.status == httplib.OK

    def ex_start_node(self, node):
        # NOTE: This method is here for backward compatibility reasons after
        # this method was promoted to be part of the standard compute API in
        # Libcloud v2.7.0
        return self.start_node(node=node)

    def ex_stop_node(self, node):
        # NOTE: This method is here for backward compatibility reasons after
        # this method was promoted to be part of the standard compute API in
        # Libcloud v2.7.0
        return self.stop_node(node=node)

    def ex_reinstall_node(self, node):
        params = {'type': 'reinstall'}
        res = self.connection.request('/metal/v1/devices/%s/actions'
                                      % (node.id),
                                      params=params, method='POST')
        return res.status == httplib.OK

    def ex_rescue_node(self, node):
        params = {'type': 'rescue'}
        res = self.connection.request('/metal/v1/devices/%s/actions'
                                      % (node.id),
                                      params=params, method='POST')
        return res.status == httplib.OK

    def ex_update_node(self, node, **kwargs):
        path = '/metal/v1/devices/%s' % node.id
        res = self.connection.request(path, params=kwargs, method='PUT')
        return res.status == httplib.OK

    def ex_get_node_bandwidth(self, node, from_time, until_time):
        path = '/metal/v1/devices/%s/bandwidth' % node.id
        params = {'from': from_time, 'until': until_time}
        return self.connection.request(path, params=params).object

    def ex_list_ip_assignments_for_node(self, node, include=''):
        path = '/metal/v1/devices/%s/ips' % node.id
        params = {'include': include}
        return self.connection.request(path, params=params).object

    def list_key_pairs(self):
        """
        List all the available SSH keys.

        :return: Available SSH keys.
        :rtype: ``list`` of :class:`.KeyPair` objects
        """
        data = self.connection.request(
            '/metal/v1/ssh-keys').object['ssh_keys']
        return list(map(self._to_key_pairs, data))

    def create_key_pair(self, name, public_key):
        """
        Create a new SSH key.

        :param      name: Key name (required)
        :type       name: ``str``

        :param      public_key: Valid public key string (required)
        :type       public_key: ``str``
        """
        params = {'label': name, 'key': public_key}
        data = self.connection.request('/metal/v1/ssh-keys',
                                       method='POST',
                                       params=params).object
        return self._to_key_pairs(data)

    def delete_key_pair(self, key):
        """
        Delete an existing SSH key.

        :param      key: SSH key (required)
        :type       key: :class:`KeyPair`
        """
        key_id = key.name
        res = self.connection.request('/metal/v1/ssh-keys/%s'
                                      % (key_id),
                                      method='DELETE')
        return res.status == httplib.NO_CONTENT

    def _to_node(self, data):
        extra = {}
        extra_keys = ['created_at', 'updated_at',
                      'userdata', 'billing_cycle', 'locked',
                      'iqn', 'locked', 'project', 'description']
        if 'state' in data:
            state = self.NODE_STATE_MAP.get(data['state'], NodeState.UNKNOWN)
        else:
            state = NodeState.UNKNOWN

        if 'ip_addresses' in data and data['ip_addresses'] is not None:
            ips = self._parse_ips(data['ip_addresses'])

        if 'operating_system' in data and data['operating_system'] is not None:
            image = self._to_image(data['operating_system'])
            extra['operating_system'] = data['operating_system'].get('name')
        else:
            image = None

        if 'plan' in data and data['plan'] is not None:
            size = self._to_size(data['plan'])
            extra['plan'] = data['plan'].get('slug')
        else:
            size = None
        if 'facility' in data:
            extra['facility'] = data['facility']

        for key in extra_keys:
            if key in data:
                extra[key] = data[key]

        node = Node(id=data['id'], name=data['hostname'], state=state,
                    public_ips=ips['public'], private_ips=ips['private'],
                    size=size, image=image, extra=extra, driver=self)
        return node

    def _to_image(self, data):
        extra = {
            'distro': data['distro'],
            'version': data['version'],
            'supported_sizes': data.get('provisionable_on', [])
        }
        return NodeImage(id=data['slug'], name=data['name'], extra=extra,
                         driver=self)

    def _to_location(self, data):
        extra = data
        return NodeLocation(id=data['id'], name=data['name'], country=None,
                            driver=self, extra=extra)

    def _to_size(self, data):
        try:
            cpus = data['specs']['cpus'][0].get('count')
        except KeyError:
            cpus = None
        regions = [region.get('href').replace('/metal/v1/facilities/', '')
                   for region in data.get('available_in', [])]
        extra = {'description': data['description'], 'line': data['line'],
                 'cpus': cpus, 'regions': regions}
        try:
            ram = int(data['specs']['memory']['total'].replace('GB', '')) * 1024  # noqa
        except KeyError:
            ram = None
        disk = None
        if data['specs'].get('drives', ''):
            disk = 0
            for disks in data['specs']['drives']:
                disk_size = disks['size'].replace('GB', '')
                if 'TB' in disk_size:
                    disk_size = float(disks['size'].replace('TB', '')) * 1000
                disk += disks['count'] * int(disk_size)
        name = "%s - %s RAM" % (data.get('name'), ram)
        price = data['pricing'].get('hour')
        return NodeSize(id=data['slug'], name=name,
                        ram=ram, disk=disk, bandwidth=0,
                        price=price, extra=extra, driver=self)

    def _to_key_pairs(self, data):
        extra = {'label': data['label'],
                 'created_at': data['created_at'],
                 'updated_at': data['updated_at']}
        return KeyPair(name=data['id'],
                       fingerprint=data['fingerprint'],
                       public_key=data['key'],
                       private_key=None,
                       driver=self,
                       extra=extra)

    def _parse_ips(self, data):
        public_ips = []
        private_ips = []
        for address in data:
            if 'address' in address and address['address'] is not None:
                if 'public' in address and address['public'] is True:
                    public_ips.append(address['address'])
                else:
                    private_ips.append(address['address'])
        return {'public': public_ips, 'private': private_ips}

    def ex_get_bgp_config_for_project(self, ex_project_id):
        path = '/metal/v1/projects/%s/bgp-config' % ex_project_id
        return self.connection.request(path).object

    def ex_get_bgp_config(self, ex_project_id=None):
        if ex_project_id:
            projects = [ex_project_id]
        elif self.project_id:
            projects = [self.project_id]
        else:
            projects = [p.id for p in self.projects]
        retval = []
        for p in projects:
            config = self.ex_get_bgp_config_for_project(p)
            if config:
                retval.append(config)
        return retval

    def ex_get_bgp_session(self, session_uuid):
        path = '/metal/v1/bgp/sessions/%s' % session_uuid
        return self.connection.request(path).object

    def ex_list_bgp_sessions_for_node(self, node):
        path = '/metal/v1/devices/%s/bgp/sessions' % node.id
        return self.connection.request(path).object

    def ex_list_bgp_sessions_for_project(self, ex_project_id):
        path = '/metal/v1/projects/%s/bgp/sessions' % ex_project_id
        return self.connection.request(path).object

    def ex_list_bgp_sessions(self, ex_project_id=None):
        if ex_project_id:
            projects = [ex_project_id]
        elif self.project_id:
            projects = [self.project_id]
        else:
            projects = [p.id for p in self.projects]
        retval = []
        for p in projects:
            retval.extend(self.ex_list_bgp_sessions_for_project(
                p)['bgp_sessions'])
        return retval

    def ex_create_bgp_session(self, node, address_family='ipv4'):
        path = '/metal/v1/devices/%s/bgp/sessions' % node.id
        params = {'address_family': address_family}
        res = self.connection.request(path, params=params, method='POST')
        return res.object

    def ex_delete_bgp_session(self, session_uuid):
        path = '/metal/v1/bgp/sessions/%s' % session_uuid
        res = self.connection.request(path, method='DELETE')
        return res.status == httplib.OK  # or res.status == httplib.NO_CONTENT

    def ex_list_events_for_node(self, node, include=None,
                                page=1, per_page=10):
        path = '/metal/v1/devices/%s/events' % node.id
        params = {
            'include': include,
            'page': page,
            'per_page': per_page
        }
        return self.connection.request(path, params=params).object

    def ex_list_events_for_project(self, project, include=None, page=1,
                                   per_page=10):
        path = '/metal/v1/projects/%s/events' % project.id
        params = {
            'include': include,
            'page': page,
            'per_page': per_page
        }
        return self.connection.request(path, params=params).object

    def ex_describe_all_addresses(self, ex_project_id=None,
                                  only_associated=False):
        if ex_project_id:
            projects = [ex_project_id]
        elif self.project_id:
            projects = [self.project_id]
        else:
            projects = [p.id for p in self.projects]
        retval = []
        for project in projects:
            retval.extend(self.ex_describe_all_addresses_for_project(
                project, only_associated))
        return retval

    def ex_describe_all_addresses_for_project(self, ex_project_id,
                                              include=None,
                                              only_associated=False):
        """
        Returns all the reserved IP addresses for this project
        optionally, returns only addresses associated with nodes.

        :param    only_associated: If true, return only the addresses
                                   that are associated with an instance.
        :type     only_associated: ``bool``

        :return:  List of IP addresses.
        :rtype:   ``list`` of :class:`dict`
        """
        path = '/metal/v1/projects/%s/ips' % ex_project_id
        params = {
            'include': include,
        }
        ip_addresses = self.connection.request(path, params=params).object
        result = [a for a in ip_addresses.get('ip_addresses', [])
                  if not only_associated or len(a.get('assignments', [])) > 0]
        return result

    def ex_describe_address(self, ex_address_id, include=None):
        path = '/metal/v1/ips/%s' % ex_address_id
        params = {
            'include': include,
        }
        result = self.connection.request(path, params=params).object
        return result

    def ex_request_address_reservation(self, ex_project_id, location_id=None,
                                       address_family='global_ipv4',
                                       quantity=1, comments='',
                                       customdata=''):
        path = '/metal/v1/projects/%s/ips' % ex_project_id
        params = {
            'type': address_family,
            'quantity': quantity,
        }
        if location_id:
            params['facility'] = location_id
        if comments:
            params['comments'] = comments
        if customdata:
            params['customdata'] = customdata
        result = self.connection.request(
            path, params=params, method='POST').object
        return result

    def ex_associate_address_with_node(self, node, address, manageable=False,
                                       customdata=''):
        path = '/metal/v1/devices/%s/ips' % node.id
        params = {
            'address': address,
            'manageable': manageable,
            'customdata': customdata
        }
        result = self.connection.request(
            path, params=params, method='POST').object
        return result

    def ex_disassociate_address(self, address_uuid, include=None):
        path = '/metal/v1/ips/%s' % address_uuid
        params = {}
        if include:
            params['include'] = include
        result = self.connection.request(
            path, params=params, method='DELETE').object
        return result

    def list_volumes(self, ex_project_id=None):
        if ex_project_id:
            return self.ex_list_volumes_for_project(
                ex_project_id=ex_project_id)

        # if project has been specified during driver initialization, then
        # return nodes for this project only
        if self.project_id:
            return self.ex_list_volumes_for_project(
                ex_project_id=self.project_id)

        # In case of Python2 perform requests serially
        if not use_asyncio():
            nodes = []
            for project in self.projects:
                nodes.extend(
                    self.ex_list_volumes_for_project(ex_project_id=project.id)
                )
            return nodes
        # In case of Python3 use asyncio to perform requests in parallel
        return self.list_resources_async('volumes')

    def ex_list_volumes_for_project(self, ex_project_id, include='plan',
                                    page=1, per_page=1000):
        params = {
            'include': include,
            'page': page,
            'per_page': per_page
        }
        data = self.connection.request(
            '/metal/v1/projects/%s/storage' % (ex_project_id),
            params=params).object['volumes']
        return list(map(self._to_volume, data))

    def _to_volume(self, data):
        return StorageVolume(id=data['id'], name=data['name'],
                             size=data['size'], driver=self,
                             extra=data)

    def create_volume(self, size, location, plan='storage_1', description='',
                      ex_project_id=None, locked=False, billing_cycle=None,
                      customdata='', snapshot_policies=None, **kwargs):
        """
        Create a new volume.

        :param size: Size of volume in gigabytes (required)
        :type size: ``int``

        :param location: Which data center to create a volume in. If
                               empty, undefined behavior will be selected.
                               (optional)
        :type location: :class:`.NodeLocation`
        :return: The newly created volume.
        :rtype: :class:`StorageVolume`
        """
        path = '/metal/v1/projects/%s/storage' % (
            ex_project_id or self.projects[0].id)
        try:
            facility = location.extra['code']
        except AttributeError:
            facility = location
        params = {
            'facility': facility,
            'plan': plan,
            'size': size,
            'locked': locked
        }
        params.update(kwargs)
        if description:
            params['description'] = description
        if customdata:
            params['customdata'] = customdata
        if billing_cycle:
            params['billing_cycle'] = billing_cycle
        if snapshot_policies:
            params['snapshot_policies'] = snapshot_policies
        data = self.connection.request(
            path, params=params, method='POST').object
        return self._to_volume(data)

    def destroy_volume(self, volume):
        """
        Destroys a storage volume.

        :param volume: Volume to be destroyed
        :type volume: :class:`StorageVolume`

        :rtype: ``bool``
        """
        path = '/metal/v1/storage/%s' % volume.id
        res = self.connection.request(path, method='DELETE')
        return res.status == httplib.NO_CONTENT

    def attach_volume(self, node, volume):
        """
        Attaches volume to node.

        :param node: Node to attach volume to.
        :type node: :class:`.Node`

        :param volume: Volume to attach.
        :type volume: :class:`.StorageVolume`

        :rytpe: ``bool``
        """
        path = '/metal/v1/storage/%s/attachments' % volume.id
        params = {
            'device_id': node.id
        }
        res = self.connection.request(path, params=params, method='POST')
        return res.status == httplib.OK

    def detach_volume(self, volume, ex_node=None, ex_attachment_id=''):
        """
        Detaches a volume from a node.

        :param volume: Volume to be detached
        :type volume: :class:`.StorageVolume`

        :param ex_attachment_id: Attachment id to be detached, if empty detach
                                        all attachments
        :type name: ``str``

        :rtype: ``bool``
        """
        path = '/metal/v1/storage/%s/attachments' % volume.id
        attachments = volume.extra['attachments']
        assert len(attachments) > 0, "Volume is not attached to any node"
        success = True
        result = None
        for attachment in attachments:
            if not ex_attachment_id or ex_attachment_id in attachment['href']:
                attachment_id = attachment['href'].split('/')[-1]
                if ex_node:
                    node_id = self.ex_describe_attachment(
                        attachment_id)['device']['href'].split('/')[-1]
                    if node_id != ex_node.id:
                        continue
                path = '/metal/v1/storage/attachments/%s' % (
                    ex_attachment_id or attachment_id)
                result = self.connection.request(path, method='DELETE')
                success = success and result.status == httplib.NO_CONTENT

        return result and success

    def create_volume_snapshot(self, volume, name=''):
        """
        Create a new volume snapshot.

        :param volume: Volume to create a snapshot for
        :type volume: class:`StorageVolume`

        :return: The newly created volume snapshot.
        :rtype: :class:`VolumeSnapshot`
        """
        path = '/metal/v1/storage/%s/snapshots' % volume.id
        res = self.connection.request(path, method='POST')
        assert res.status == httplib.ACCEPTED
        return volume.list_snapshots()[-1]

    def destroy_volume_snapshot(self, snapshot):
        """
        Delete a volume snapshot

        :param snapshot: volume snapshot to delete
        :type snapshot: class:`VolumeSnapshot`

        :rtype: ``bool``
        """
        volume_id = snapshot.extra['volume']['href'].split('/')[-1]
        path = '/metal/v1/storage/%s/snapshots/%s' % (volume_id, snapshot.id)
        res = self.connection.request(path, method='DELETE')
        return res.status == httplib.NO_CONTENT

    def list_volume_snapshots(self, volume, include=''):
        """
        List snapshots for a volume.

        :param volume: Volume to list snapshots for
        :type volume: class:`StorageVolume`

        :return: List of volume snapshots.
        :rtype: ``list`` of :class: `VolumeSnapshot`
        """
        path = '/metal/v1/storage/%s/snapshots' % volume.id
        params = {}
        if include:
            params['include'] = include
        data = self.connection.request(path, params=params).object['snapshots']
        return list(map(self._to_volume_snapshot, data))

    def _to_volume_snapshot(self, data):
        created = datetime.datetime.strptime(
            data['created_at'], "%Y-%m-%dT%H:%M:%S")
        return VolumeSnapshot(id=data['id'],
                              name=data['id'],
                              created=created,
                              state=data['status'],
                              driver=self, extra=data)

    def ex_modify_volume(self, volume, description=None, size=None,
                         locked=None, billing_cycle=None,
                         customdata=None):
        path = '/metal/v1/storage/%s' % volume.id
        params = {}
        if description:
            params['description'] = description
        if size:
            params['size'] = size
        if locked is not None:
            params['locked'] = locked
        if billing_cycle:
            params['billing_cycle'] = billing_cycle
        res = self.connection.request(path, params=params, method='PUT')
        return self._to_volume(res.object)

    def ex_restore_volume(self, snapshot):
        volume_id = snapshot.extra['volume']['href'].split('/')[-1]
        ts = snapshot.extra['timestamp']
        path = '/metal/v1/storage/%s/restore?restore_point=%s' % (
            volume_id, ts)
        res = self.connection.request(path, method='POST')
        return res.status == httplib.NO_CONTENT

    def ex_clone_volume(self, volume, snapshot=None):
        path = '/metal/v1/storage/%s/clone' % volume.id
        if snapshot:
            path += '?snapshot_timestamp=%s' % snapshot.extra['timestamp']
        res = self.connection.request(path, method='POST')
        return res.status == httplib.NO_CONTENT

    def ex_describe_volume(self, volume_id):
        path = '/metal/v1/storage/%s' % volume_id
        data = self.connection.request(path).object
        return self._to_volume(data)

    def ex_describe_attachment(self, attachment_id):
        path = '/metal/v1/storage/attachments/%s' % attachment_id
        data = self.connection.request(path).object
        return data


class Project(object):
    def __init__(self, project):
        self.id = project.get('id')
        self.name = project.get('name')
        self.extra = {}
        self.extra['max_devices'] = project.get('max_devices')
        self.extra['payment_method'] = project.get('payment_method')
        self.extra['created_at'] = project.get('created_at')
        self.extra['credit_amount'] = project.get('credit_amount')
        self.extra['devices'] = project.get('devices')
        self.extra['invitations'] = project.get('invitations')
        self.extra['memberships'] = project.get('memberships')
        self.extra['href'] = project.get('href')
        self.extra['members'] = project.get('members')
        self.extra['ssh_keys'] = project.get('ssh_keys')

    def __repr__(self):
        return (('<Project: id=%s, name=%s>') %
                (self.id, self.name))



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import pytest


def pytest_configure(config):
    """Check that secrets.py is valid"""

    this_dir = os.path.abspath(os.path.split(__file__)[0])
    secrets_current = os.path.join(this_dir, 'secrets.py')
    secrets_dist = os.path.join(this_dir, 'secrets.py-dist')

    if not os.path.isfile(secrets_current):
        print("Missing " + secrets_current)
        print("Maybe you forgot to copy it from -dist:")
        print("cp libcloud/test/secrets.py-dist libcloud/test/secrets.py")
        pytest.exit('')

    mtime_current = os.path.getmtime(secrets_current)
    mtime_dist = os.path.getmtime(secrets_dist)

    if mtime_dist > mtime_current:
        print("It looks like test/secrets.py file is out of date.")
        print("Please copy the new secrets.py-dist file over otherwise" +
              " tests might fail")
        pytest.exit('')
